"""Automation tests for sssctl analyze

:requirement: sssctl analyze
:casecomponent: sssd
:subsystemteam: sst_idm_sssd
:upstream: yes
:status: approved
"""
import pytest
import re
from sssd.testlib.common.utils import sssdTools
from sssd.testlib.common.ssh2_python import check_login_client
from constants import ds_instance_name


def analyze(multihost, req_arg, op_arg=None):
    """
    Execute sssctl analyze command with given action and options
    and return the exit status as well as command output
        :param str req_arg: command action to perform, values could be
        'list' or 'show' or 'show <num>'
        : param str arg: optional argument string. values could be
        '--source <log_source>' or '--logdir <dir_path>' or '--help'
    """
    if op_arg is None:
        op_arg = ''
    str_cmd = f'sssctl analyze {op_arg} request {req_arg}'
    ss_cmd = multihost.client[0].run_command(str_cmd, raiseonerr=False)
    return ss_cmd.returncode, ss_cmd.stdout_text


@pytest.mark.usefixtures('setup_sssd_krb', 'create_posix_usersgroups')
@pytest.mark.analyze
@pytest.mark.tier1_4
class TestSssctlAnalyze(object):
    """ sssctl analyze test suite """
    @pytest.mark.converted('test_sssctl.py', 'test_sssctl__analyze_list')
    def test_analyze_list(self, multihost, backupsssdconf):
        """
        :title: sssctl analyze list to show captured nss related
         requests from sssd log
        :id: 95e18ae1-6c4a-4baa-8202-fe33fe82bdec
        :description: sssctl analyze request list is able to capture the user
         and group related requests raised when commands like id and getent
         are executed
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=1294670
        :steps:
          1. Configure sssd to authenticate against directory server
          2. Enable debug_level to 9 in the 'nss', 'pam' and domain section
          3. Restart SSSD with cleared cache
          4. Fetch user and group information using 'id' and 'getent' tools
          5. Run 'sssctl analyze request list'
          6. Check with sssctl analyze is listing id and getent instances
          7. sssctl analyze with subcmd 'show' and request number is listing
             logs related to that number only
        :expectedresults:
          1. Should succeed
          2. Should succeed
          3. Should succeed
          4. Should succeed
          5. Should succeed
          6. Should succeed
          7. Should succeed
        """
        tools = sssdTools(multihost.client[0])
        dm_sec = ['nss', 'pam']
        sssd_params = {'debug_level': '9'}
        for sec_op in dm_sec:
            tools.sssd_conf(sec_op, sssd_params, action='update')
        multihost.client[0].service_sssd('start')
        tools.clear_sssd_cache()
        g_cmd = f'getent group ldapusers@{ds_instance_name}'
        multihost.client[0].run_command(g_cmd, raiseonerr=False)
        i_cmd = f'id foo1@{ds_instance_name}'
        multihost.client[0].run_command(i_cmd, raiseonerr=False)
        for act_op in ['list', 'list -v']:
            _, stdout = analyze(multihost, act_op)
            assert all(ptn in stdout for ptn in [' id', 'getent'])
        tools.clear_sssd_cache()
        g_cmd = f'getent passwd foo1@{ds_instance_name}'
        multihost.client[0].run_command(g_cmd, raiseonerr=False)
        for act_op in ['list', 'list -v']:
            _, stdout = analyze(multihost, act_op)
            assert all(ptn in stdout for ptn in ['CID #1', 'getent'])

    @pytest.mark.converted('test_sssctl.py', 'test_sssctl__analyze_non_default_log_location')
    def test_analyze_diff_log_location(self, multihost, backupsssdconf):
        """
        :title: sssctl analyze able to parse sssd logs from non-default
         location
        :description: sssctl analyze should be able to parse the sssd logs
         from different location or logs from other host
        :id: d297b394-3502-4ade-a5a5-5fb4c4333645
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=1294670
                   https://github.com/SSSD/sssd/issues/6298
        :steps:
          1. Configure sssd to authenticate against directory server
          2. Enable debug_level to 9 in the 'nss', 'pam' and domain section
          3. Restart SSSD with cleared cache
          4. Fetch user as well as  information using 'id' and 'groups' tools
          5. Log in as user via ssh
          6. Copy sssd logs to a different location
          7. Stop sssd and remove conf, logs and cache
          8. Confirm --logdir allows analyze to parse logs from that location
        :expectedresults:
          1. Should succeed
          2. Should succeed
          3. Should succeed
          4. Should succeed
          5. Should succeed
          6. Should succeed
          7. No sssd running or configured
          8. Should succeed
        """
        tools = sssdTools(multihost.client[0])
        dm_sec = ['nss', 'pam']
        sssd_params = {'debug_level': '9'}
        for sec_op in dm_sec:
            tools.sssd_conf(sec_op, sssd_params, action='update')
        tools.clear_sssd_cache()
        user = f'foo1@{ds_instance_name}'
        i_cmd = f'id {user}'
        multihost.client[0].run_command(i_cmd, raiseonerr=False)
        check_login_client(multihost, user, 'Secret123')
        cp_cmd = 'cp -r /var/log/sssd /tmp/'
        multihost.client[0].run_command(cp_cmd, raiseonerr=False)
        multihost.client[0].service_sssd('stop')
        cp_cmd = 'rm -f /etc/sssd/sssd.conf'
        multihost.client[0].run_command(cp_cmd, raiseonerr=False)
        tools.remove_sss_cache('/var/log/sssd/')
        tools.remove_sss_cache('/var/lib/sss/db/')
        ss_op = 'show 1 --pam'
        log_dir = '--logdir /tmp/sssd/'
        _, stdout = analyze(multihost, ss_op, log_dir)
        pam_cmds = ['SSS_PAM_AUTHENTICATE', 'SSS_PAM_ACCT_MGMT',
                    'SSS_PAM_SETCRED']
        for pam_auth in pam_cmds:
            assert pam_auth in stdout
        for act_op in ['list', 'list -v']:
            _, stdout = analyze(multihost, act_op, log_dir)
            assert 'id' in stdout
        for act_op in ['list --pam', 'list -v --pam']:
            _, stdout = analyze(multihost, act_op, log_dir)
            assert 'sshd' in stdout

    @pytest.mark.converted('test_sssctl.py', 'test_sssctl__analyze_pam_logs')
    def test_analyze_pam_logs(self, multihost, backupsssdconf):
        """
        :title: sssctl analyze to parse pam requests from logs
        :id: 7fcd03b6-7f6f-4f39-96f8-45e0cb2d8c20
        :description: sssctl analyze request should able to parse and return
         authentication logs
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=1294670
        :steps:
          1. Configure sssd to authenticate against directory server
          2. Enable debug_level to 9 in the 'nss', 'pam' and domain section
          3. Restart SSSD with cleared cache
          4. Log in as a user using ssh
          5. Confirm --pam option is showing login related logs
        :expectedresults:
          1. Should succeed
          2. Should succeed
          3. Should succeed
          4. Should succeed
          5. Should succeed
        """
        tools = sssdTools(multihost.client[0])
        multihost.client[0].service_sssd('stop')
        tools.remove_sss_cache('/var/lib/sss/db/')
        tools.remove_sss_cache('/var/log/sssd/')
        dm_sec = ['nss', 'pam']
        sssd_params = {'debug_level': '9'}
        for sec_op in dm_sec:
            tools.sssd_conf(sec_op, sssd_params, action='update')
        multihost.client[0].service_sssd('start')
        user = f'foo1@{ds_instance_name}'
        check_login_client(multihost, user, 'Secret123')
        _, stdout = analyze(multihost, 'show 1 --pam')
        assert 'CID #1' in stdout
        pam_cmds = ['SSS_PAM_AUTHENTICATE', 'SSS_PAM_ACCT_MGMT',
                    'SSS_PAM_SETCRED']
        for pam_auth in pam_cmds:
            assert pam_auth in stdout

    @pytest.mark.converted('test_sssctl.py', 'test_sssctl__analyze_tevent_id')
    def test_analyze_tevent_id(self, multihost, backupsssdconf):
        """
        :title: sssctl analyze to parse tevent chain IDs from logs
        :id: f748766c-0177-4306-9e7f-816586734e14
        :description: sssctl analyze should able to parse tevent chain
         IDs from responder logs
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=2013259
        :steps:
          1. Configure sssd to authenticate against directory server
          2. Enable debug_level to 9 in the 'nss', 'pam' and domain section
          3. Restart SSSD with cleared cache
          4. Log in as a user using ssh
          5. Confirm tevent chain IDs(RID) is showing in logs
        :expectedresults:
          1. Should succeed
          2. Should succeed
          3. Should succeed
          4. Should succeed
          5. Should succeed
        """
        tools = sssdTools(multihost.client[0])
        dm_sec = ['nss', 'pam']
        sssd_params = {'debug_level': '9'}
        for sec_op in dm_sec:
            tools.sssd_conf(sec_op, sssd_params, action='update')
        tools.clear_sssd_cache()
        i_cmd = f'id foo1@{ds_instance_name}'
        multihost.client[0].run_command(i_cmd, raiseonerr=False)
        user = f'foo1@{ds_instance_name}'
        check_login_client(multihost, user, 'Secret123')
        _, stdout = analyze(multihost, 'show 1 --pam')
        assert all(ptn in stdout for ptn in ['RID#', user])

    @pytest.mark.converted('test_sssctl.py', 'test_sssctl__analyze_parse_child_logs')
    def test_analyze_parse_child_logs(self, multihost, backupsssdconf):
        """
        :title: sssctl analyze to parse child logs from logs
        :id: 0f009b2e-420f-40f4-ab37-e224a6607812
        :description: sssctl analyze should able to parse child logs
         from  logs
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=2013260
        :steps:
          1. Configure sssd to authenticate against directory server
          2. Enable debug_level to 9 in the 'nss', 'pam' and domain section
          3. Restart SSSD with cleared cache
          4. Log in as a user using ssh
          5. Confirm child krb logs  parsed
          6. Fail log in with wrong credentials
          7. Confirm parsed child logs show error message
        :expectedresults:
          1. Should succeed
          2. Should succeed
          3. Should succeed
          4. Should succeed
          5. Should succeed
          6. Should succeed
          7. Should succeed
        """
        tools = sssdTools(multihost.client[0])
        dm_sec = ['nss', 'pam']
        sssd_params = {'debug_level': '9'}
        for sec_op in dm_sec:
            tools.sssd_conf(sec_op, sssd_params, action='update')
        tools.clear_sssd_cache()
        user = f'foo1@{ds_instance_name}'
        try:
            check_login_client(multihost, user, 'Secret123')
        except Exception:
            _, stdout = analyze(multihost, 'show --pam --child 1')
            assert 'Preauthentication failed' in stdout
            pytest.fail(f"{user} failed to login")
        _, stdout = analyze(multihost, 'show --pam --child 1')
        err = 'sss_child_krb5_trace_cb'
        assert all(ptn in stdout for ptn in [err, user])
        tools.clear_sssd_cache()
        try:
            check_login_client(multihost, user, 'Secret123')
        except Exception:
            _, stdout = analyze(multihost, 'show --pam --child 1')
            assert re.findall(r"RID#[0-9]*] Received error code", stdout)

    @pytest.mark.converted('test_sssctl.py', 'test_sssctl__analyze_root_privileges')
    @staticmethod
    def test_non_root_privileged(multihost, localusers):
        """
        :title: SSSD: `sssctl analyze` command shouldn't require 'root' privileged
        :id: 51a69d4e-7ae4-11ed-95a5-845cf3eff344
        :description: `sssctl analyze` command shouldn't require
            'root' privileged when run with `--logdir`
            pointing to otherwise accessible files.
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=2142960
                   https://bugzilla.redhat.com/show_bug.cgi?id=2142794
                   https://bugzilla.redhat.com/show_bug.cgi?id=2142961
        :steps:
          1. Create directory
          2. Copy logs to above directory
          3. Change ownership to a local user
          4. Try sssctl analyze command pointing to above
            directory with root and loacluser
        :expectedresults:
          1. Should succeed
          2. Should succeed
          3. Should succeed
          4. sssctl analyze outputs provided for user root
            and user5000 are the same
        """
        # Create directory
        # Copy logs to above directory
        # Change ownership to a local user
        multihost.client[0].run_command("rm -vfr /tmp/sssd", raiseonerr=False)
        for command in ["mkdir /tmp/sssd",
                        "cp -vf /var/log/sssd/* /tmp/sssd",
                        "chown user5000 /tmp/sssd/",
                        "chown user5000 /tmp/sssd/*"]:
            multihost.client[0].run_command(command)
        # sssctl analyze command with root
        cmd_root = multihost.client[0].run_command("sssctl analyze --logdir /tmp/sssd")
        # sssctl analyze command with non root user
        cmd_user500 = multihost.client[0].run_command("runuser -l user5000 -c "
                                                      "'sssctl analyze --logdir /tmp/sssd'")
        multihost.client[0].run_command("rm -vfr /tmp/sssd")
        assert cmd_root.returncode == 0
        assert cmd_user500.returncode == 0
        # sssctl analyze command output should same for both
        # root and non root user
        assert cmd_root.stdout_text == cmd_user500.stdout_text



"""Automation for cache performance tests ported from bash

:requirement: SSSD Memory cache Performance
:casecomponent: sssd
:subsystemteam: sst_idm_sssd
:upstream: yes
:status: approved
"""

import pytest
import time
from constants import ds_instance_name
from sssd.testlib.common.utils import sssdTools
from sssd.testlib.common.ssh2_python import check_login_client_bool


@pytest.fixture(scope='class')
def custom_setup(multihost, setup_sssd, create_posix_usersgroups):
    """ Added neccessary sssd domain parameters """
    tools = sssdTools(multihost.client[0])
    sssd_params = {'services': "nss, pam"}
    tools.sssd_conf('sssd', sssd_params)
    domain_section = f'domain/{ds_instance_name}'
    domain_params = {'use_fully_qualified_names': 'False',
                     'override_homedir': '/home/%u'}
    tools.sssd_conf(domain_section, domain_params)
    tools.clear_sssd_cache()


@pytest.mark.tier2
@pytest.mark.cache_performance
@pytest.mark.usefixtures('custom_setup')
class TestCacheTesting():
    """
    This is test case class for cache_performance testing suite

    Test SSSD Cache existence and presence of updated information
    on the basis of several User and Group Modifications.
    Test for correct cache presence in case of user and group lookups.
    Test for correct cache presence in case of expiration of user and group entries.
    Test for correct cache presence in case of user authentication.
    Test for correct cache presence in case of variety of SSSD Configurations.
    Test for correct cache presence in case of modification & deletion of user entries.
    """
    @pytest.mark.converted('test_sss_cache.py', 'test_sss_cache__check_timestamp_value_in_ldb')
    @staticmethod
    def test_0001_Verify_Timestamp_Cache_Exists(multihost):
        """
        :title: IDM-SSSD-TC: ldap_provider: cache_performance: Verify the existence Of Timestamp Cache
        :id: d58f6bc7-97bd-459c-9871-f3e0ae6c449e
        :setup:
          1. Clear the sssd cache and restart sssd.
        :steps:
          1. Execute getent to fetch user details.
          2. Check if timestamps cache file exits.
        :expectedresults:
          1. User details should be successfully fetched.
          2. Cache file should be present.
        """
        client = sssdTools(multihost.client[0])
        client.clear_sssd_cache()
        client.get_getent_passwd('foo1')
        file = f"/var/lib/sss/db/timestamps_{ds_instance_name}.ldb"
        list_cmd = f"ls -ld {file}"
        cmd = multihost.client[0].run_command(list_cmd, raiseonerr=False)
        assert cmd.returncode == 0, f"Could not find timestamp cache file {file}"

    @pytest.mark.converted('test_sss_cache.py', 'test_sss_cache__check_timestamp_value_in_ldb_when_fully_qualified_names_enabled')
    @staticmethod
    def test_0002_Verify_Cache_on_User_Lookup(multihost):
        """
        :title: IDM-SSSD-TC: ldap_provider: cache_performance: Verify ldb cache updates on user lookup
        :id: e71e1a88-7521-406c-9630-7201362b0587
        :steps:
          1. Execute the ldbsearch command to get cache db output
          2. Execute the ldbsearch command to get timestamps db output
        :expectedresults:
          1. Command executes and cache file should be present.
          2. Command executes and cache file should be present.
        """
        ldb_cmd = f"ldbsearch -H /var/lib/sss/db/cache_{ds_instance_name}.ldb \
                    -b 'name=foo1@{ds_instance_name},cn=users,cn={ds_instance_name},cn=sysdb' | tee /tmp/output"
        ldb_cmd2 = f"ldbsearch -H /var/lib/sss/db/timestamps_{ds_instance_name}.ldb \
                    -b 'name=foo1@{ds_instance_name},cn=users,cn={ds_instance_name},cn=sysdb'"
        cmd = multihost.client[0].run_command(ldb_cmd, raiseonerr=False)
        cmd2 = multihost.client[0].run_command(ldb_cmd2, raiseonerr=False)
        assert cmd.returncode == 0, f'{ldb_cmd} did not execute successfully'
        assert cmd2.returncode == 0, f'{ldb_cmd2} did not execute successfully'

    @staticmethod
    def test_0003_Expire_User_Entries_and_Verify_Updates(multihost):
        """
        :title: IDM-SSSD-TC: ldap_provider: cache_performance: Expire user entries in cache and verify the updates
        :id: 4a2d7d2a-74cc-4d14-ae4b-d2693f3c5181
        :steps:
          1. Invalidate the existing cache
          2. Execute the ldbsearch command and check for dataExpireTimestamp in the cache db output
          3. Execute the ldbsearch command and check for dataExpireTimestamp in the timestamps db output
        :expectedresults:
          1. Cache sucessfully get invalidated.
          2. dataExpireTimestamp should be present in cache db output.
          3. dataExpireTimestamp should be present in timestamps db output.
        """
        invalidate_cache = "sss_cache -E"
        cmd = multihost.client[0].run_command(invalidate_cache, raiseonerr=False)
        ldb_cmd1 = f"ldbsearch -H /var/lib/sss/db/cache_{ds_instance_name}.ldb \
                    -b 'name=foo1@{ds_instance_name},cn=users,cn={ds_instance_name},cn=sysdb' > /tmp/file_ldb1"
        ldb_cmd2 = f"ldbsearch -H /var/lib/sss/db/timestamps_{ds_instance_name}.ldb \
                    -b 'name=foo1@{ds_instance_name},cn=users,cn={ds_instance_name},cn=sysdb' > /tmp/file_ldb2"
        cmd1 = multihost.client[0].run_command(ldb_cmd1, raiseonerr=False)
        cmd2 = multihost.client[0].run_command(ldb_cmd2, raiseonerr=False)
        cmd1_output = multihost.client[0].get_file_contents('/tmp/file_ldb1').decode('utf-8')
        cmd2_output = multihost.client[0].get_file_contents('/tmp/file_ldb2').decode('utf-8')
        assert cmd.returncode == 0, f'{invalidate_cache} did not execute successfully'
        assert cmd1.returncode == 0, f'{ldb_cmd1} did not execute successfully'
        assert "dataExpireTimestamp: 1\n" in cmd1_output, "dataExpireTimestamp not found in /tmp/file_ldb1"
        assert cmd2.returncode == 0, f'{ldb_cmd2} did not execute successfully'
        assert "dataExpireTimestamp: 1\n" in cmd2_output, "dataExpireTimestamp not found in /tmp/file_ldb2"

    @staticmethod
    def test_0004_Refresh_User_Entries_After_Expiry(multihost):
        """
        :title: IDM-SSSD-TC: ldap_provider: cache_performance: Refresh user entries after expiry and \
            verify the cache updates
        :id: 05e1a77b-bb9d-422b-ba4e-3b34f0752f65
        :steps:
          1. Execute getent to fetch user details.
          2. Execute the ldbsearch command and check for dataExpireTimestamp in the cache db output
          3. Execute the ldbsearch command and check for dataExpireTimestamp in the timestamps db output
        :expectedresults:
          1. User details should be successfully fetched.
          2. dataExpireTimestamp should be present in cache db output.
          3. dataExpireTimestamp should not be present in timestamps db output.
        """
        client = sssdTools(multihost.client[0])
        client.get_getent_passwd('foo1')
        ldb_cmd = f"ldbsearch -H /var/lib/sss/db/cache_{ds_instance_name}.ldb \
                    -b 'name=foo1@{ds_instance_name},cn=users,cn={ds_instance_name},cn=sysdb' > /tmp/file_ldb"
        ldb_cmd2 = f"ldbsearch -H /var/lib/sss/db/timestamps_{ds_instance_name}.ldb \
                    -b 'name=foo1@{ds_instance_name},cn=users,cn={ds_instance_name},cn=sysdb' > /tmp/file_ldb2"
        cmd = multihost.client[0].run_command(ldb_cmd, raiseonerr=False)
        cmd2 = multihost.client[0].run_command(ldb_cmd2, raiseonerr=False)
        cmd_output = multihost.client[0].get_file_contents('/tmp/file_ldb').decode('utf-8')
        cmd2_output = multihost.client[0].get_file_contents('/tmp/file_ldb2').decode('utf-8')
        assert cmd.returncode == 0, f'{ldb_cmd} did not execute successfully'
        assert "dataExpireTimestamp: 1\n" in cmd_output, "dataExpireTimestamp not found in /tmp/file_ldb"
        assert cmd2.returncode == 0, f'{ldb_cmd2} did not execute successfully'
        assert "dataExpireTimestamp: 1\n" not in cmd2_output, "dataExpireTimestamp found in /tmp/file_ldb2"

    @staticmethod
    def test_0005_Expire_User_Entries_ans_Run_User_Auth(multihost):
        """
        :title: IDM-SSSD-TC: ldap_provider: cache_performance: Expire entries in cache and run user auth
        :id: 70befce4-fde6-4cbd-bf7e-625e19c4dbb7
        :steps:
          1. Invalidate the existing cache
          2. Authenticate the user foo1 from the client
          3. Execute the ldbsearch command and check for dataExpireTimestamp in the cache db output
          4. Execute the ldbsearch command and check for dataExpireTimestamp in the timestamps db output
        :expectedresults:
          1. Cache sucessfully get invalidated.
          2. User foo1 should be able to successfully login
          3. dataExpireTimestamp should be present in cache db output.
          4. dataExpireTimestamp should not be present in timestamps db output.
        """
        invalidate_cache = "sss_cache -E"
        cmd = multihost.client[0].run_command(invalidate_cache, raiseonerr=False)
        ssh = check_login_client_bool(multihost, 'foo1', 'Secret123')
        ldb_cmd = f"ldbsearch -H /var/lib/sss/db/cache_{ds_instance_name}.ldb \
                    -b 'name=foo1@{ds_instance_name},cn=users,cn={ds_instance_name},cn=sysdb' > /tmp/file_ldb"
        ldb_cmd2 = f"ldbsearch -H /var/lib/sss/db/timestamps_{ds_instance_name}.ldb \
                    -b 'name=foo1@{ds_instance_name},cn=users,cn={ds_instance_name},cn=sysdb' > /tmp/file_ldb2"
        cmd1 = multihost.client[0].run_command(ldb_cmd, raiseonerr=False)
        cmd2 = multihost.client[0].run_command(ldb_cmd2, raiseonerr=False)
        cmd1_output = multihost.client[0].get_file_contents('/tmp/file_ldb').decode('utf-8')
        cmd2_output = multihost.client[0].get_file_contents('/tmp/file_ldb2').decode('utf-8')
        assert ssh, 'foo1 user is unable to login'
        assert cmd.returncode == 0, f'{invalidate_cache} did not execute successfully'
        assert cmd1.returncode == 0, f'{ldb_cmd} did not execute successfully'
        assert "dataExpireTimestamp: 1\n" in cmd1_output, "dataExpireTimestamp not found in /tmp/file_ldb"
        assert cmd2.returncode == 0, f'{ldb_cmd2} did not execute successfully'
        assert "dataExpireTimestamp: 1\n" not in cmd2_output, "dataExpireTimestamp found in /tmp/file_ldb2"

    @staticmethod
    def test_0006_Set_refresh_expired_interval_to_40(multihost, backupsssdconf):
        """
        :title: IDM-SSSD-TC: ldap_provider: cache_performance: Set refresh expired interval to 40 and \
            verify user record updates
        :id: 329e6aa3-5b97-432b-bd2e-81f6c543e3f5
        :setup:
          1. Set the refresh_expired_interval to 40 in the sssd.conf
        :steps:
          1. Execute getent to fetch user details.
          2. Invalidate the existing cache
          3. Execute the ldbsearch command and check for dataExpireTimestamp in the cache db output
          4. Execute the ldbsearch command and check for dataExpireTimestamp in the timestamps db output
        :expectedresults:
          1. User details should be successfully fetched.
          2. Cache sucessfully get invalidated.
          3. dataExpireTimestamp should be present in cache db output.
          4. dataExpireTimestamp should not be present in timestamps db output.
        """
        client = sssdTools(multihost.client[0])
        domain_section = f'domain/{ds_instance_name}'
        domain_params = {'entry_cache_timeout': 160,
                         'refresh_expired_interval': 40
                         }
        client.sssd_conf(domain_section, domain_params)
        client.clear_sssd_cache()
        client.get_getent_passwd('foo1')

        invalidate_cache = "sss_cache -E"
        cmd = multihost.client[0].run_command(invalidate_cache, raiseonerr=False)
        time.sleep(40)
        ldb_cmd = f"ldbsearch -H /var/lib/sss/db/cache_{ds_instance_name}.ldb \
                    -b 'name=foo1@{ds_instance_name},cn=users,cn={ds_instance_name},cn=sysdb' > /tmp/file_ldb"
        ldb_cmd2 = f"ldbsearch -H /var/lib/sss/db/timestamps_{ds_instance_name}.ldb \
                    -b 'name=foo1@{ds_instance_name},cn=users,cn={ds_instance_name},cn=sysdb' > /tmp/file_ldb2"
        cmd1 = multihost.client[0].run_command(ldb_cmd, raiseonerr=False)
        cmd2 = multihost.client[0].run_command(ldb_cmd2, raiseonerr=False)
        cmd1_output = multihost.client[0].get_file_contents('/tmp/file_ldb').decode('utf-8')
        cmd2_output = multihost.client[0].get_file_contents('/tmp/file_ldb2').decode('utf-8')
        assert cmd.returncode == 0, f'{invalidate_cache} did not execute successfully'
        assert cmd1.returncode == 0, f'{ldb_cmd} did not execute successfully'
        assert "dataExpireTimestamp: 1\n" in cmd1_output, "dataExpireTimestamp not found in /tmp/file_ldb"
        assert cmd2.returncode == 0, f'{ldb_cmd2} did not execute successfully'
        assert "dataExpireTimestamp: 1\n" not in cmd2_output, "dataExpireTimestamp found in /tmp/file_ldb2"

    @staticmethod
    def test_0007_Set_use_fully_qualified_names_to_true(multihost, backupsssdconf):
        """
        :title: IDM-SSSD-TC: ldap_provider: cache_performance: Set use fully qualified names to true and \
            verify cache updates
        :id: 31946f7a-2f7a-466c-ac87-0e9741bd0c80
        :setup:
          1. Set the use_fully_qualified_names to True in the sssd.conf
        :steps:
          1. Execute getent to fetch user details.
          2. Execute the ldbsearch command to get cache db output
          3. Execute the ldbsearch command to get timestamps db output
        :expectedresults:
          1. User details should be successfully fetched.
          2. Command executes and cache file should be present.
          3. Command executes and cache file should be present.
        """
        client = sssdTools(multihost.client[0])
        domain_section = f'domain/{ds_instance_name}'
        domain_params = {'use_fully_qualified_names': True}
        client.sssd_conf(domain_section, domain_params)
        client.clear_sssd_cache()
        client.get_getent_passwd('foo1@example1')
        ldb_cmd = f"ldbsearch -H /var/lib/sss/db/cache_{ds_instance_name}.ldb \
                    -b 'name=foo1@{ds_instance_name},cn=users,cn={ds_instance_name},cn=sysdb' | tee /tmp/output"
        ldb_cmd2 = f"ldbsearch -H /var/lib/sss/db/timestamps_{ds_instance_name}.ldb \
                    -b 'name=foo1@{ds_instance_name},cn=users,cn={ds_instance_name},cn=sysdb'"
        cmd = multihost.client[0].run_command(ldb_cmd, raiseonerr=False)
        cmd2 = multihost.client[0].run_command(ldb_cmd2, raiseonerr=False)
        assert cmd.returncode == 0, f'{ldb_cmd} did not execute successfully'
        assert cmd2.returncode == 0, f'{ldb_cmd2} did not execute successfully'

    @staticmethod
    def test_0008_Set_case_sensitive_to_false(multihost, backupsssdconf):
        """
        :title: IDM-SSSD-TC: ldap_provider: cache_performance: Set case sensitive to false and verify cache updates
        :id: fd1dc840-3908-4eba-81ec-ae492335b8c6
        :setup:
          1. Set the case_sensitive to False in the sssd.conf
        :steps:
          1. Execute getent to fetch user details.
          2. Execute the ldbsearch command to get cache db output
          3. Execute the ldbsearch command to get timestamps db output
        :expectedresults:
          1. User details should be successfully fetched.
          2. Command executes and cache file should be present.
          3. Command executes and cache file should be present.
        """
        client = sssdTools(multihost.client[0])
        domain_section = f'domain/{ds_instance_name}'
        domain_params = {'case_sensitive': False}
        client.sssd_conf(domain_section, domain_params)
        client.clear_sssd_cache()
        client.get_getent_passwd('FOO1@EXAMPLE1')
        ldb_cmd = f"ldbsearch -H /var/lib/sss/db/cache_{ds_instance_name}.ldb \
                    -b 'name=foo1@{ds_instance_name},cn=users,cn={ds_instance_name},cn=sysdb' | tee /tmp/output"
        ldb_cmd2 = f"ldbsearch -H /var/lib/sss/db/timestamps_{ds_instance_name}.ldb \
                    -b 'name=foo1@{ds_instance_name},cn=users,cn={ds_instance_name},cn=sysdb'"
        cmd = multihost.client[0].run_command(ldb_cmd, raiseonerr=False)
        cmd2 = multihost.client[0].run_command(ldb_cmd2, raiseonerr=False)
        assert cmd.returncode == 0, f'{ldb_cmd} did not execute successfully'
        assert cmd2.returncode == 0, f'{ldb_cmd2} did not execute successfully'

    @staticmethod
    def test_0009_Verify_ldb_cache_updates_on_group_lookup(multihost):
        """
        :title: IDM-SSSD-TC: ldap_provider: cache_performance: Verify ldb cache updates on group lookup
        :id: 8af50374-bc7f-44f2-b381-b8c28f7a2f52
        :steps:
          1. Execute getent to fetch group details.
          2. Execute the ldbsearch command to get cache db output
          3. Execute the ldbsearch command to get timestamps db output
        :expectedresults:
          1. Group details should be successfully fetched.
          2. Command executes and cache file should be present.
          3. Command executes and cache file should be present.
        """
        client = sssdTools(multihost.client[0])
        client.get_getent_group('ldapusers')
        ldb_cmd = f"ldbsearch -H /var/lib/sss/db/cache_{ds_instance_name}.ldb \
                    -b 'name=ldapusers@{ds_instance_name},cn=groups,cn={ds_instance_name},cn=sysdb' | tee /tmp/output"
        ldb_cmd2 = f"ldbsearch -H /var/lib/sss/db/timestamps_{ds_instance_name}.ldb \
                    -b 'name=ldapusers@{ds_instance_name},cn=groups,cn={ds_instance_name},cn=sysdb'"
        cmd = multihost.client[0].run_command(ldb_cmd, raiseonerr=False)
        cmd2 = multihost.client[0].run_command(ldb_cmd2, raiseonerr=False)
        assert cmd.returncode == 0, f'{ldb_cmd} did not execute successfully'
        assert cmd2.returncode == 0, f'{ldb_cmd2} did not execute successfully'

    @staticmethod
    def test_0010_Expire_group_record_in_cache(multihost):
        """
        :title: IDM-SSSD-TC: ldap_provider: cache_performance: Expire group record in cache and verify the updates
        :id: 39bc2b36-8950-4d74-84a7-8f701af09456
        :steps:
          1. Invalidate the existing cache
          2. Execute the ldbsearch command and check for dataExpireTimestamp in the cache db output
          3. Execute the ldbsearch command and check for dataExpireTimestamp in the timestamps db output
        :expectedresults:
          1. Cache sucessfully get invalidated.
          2. dataExpireTimestamp should be present in cache db output.
          3. dataExpireTimestamp should be present in timestamps db output.
        """
        invalidate_cache = "sss_cache -E"
        cmd = multihost.client[0].run_command(invalidate_cache, raiseonerr=False)
        ldb_cmd1 = f"ldbsearch -H /var/lib/sss/db/cache_{ds_instance_name}.ldb \
                    -b 'name=ldapusers@{ds_instance_name},cn=groups,cn={ds_instance_name},cn=sysdb' > /tmp/file_ldb1"
        ldb_cmd2 = f"ldbsearch -H /var/lib/sss/db/timestamps_{ds_instance_name}.ldb \
                    -b 'name=ldapusers@{ds_instance_name},cn=groups,cn={ds_instance_name},cn=sysdb' > /tmp/file_ldb2"
        cmd1 = multihost.client[0].run_command(ldb_cmd1, raiseonerr=False)
        cmd2 = multihost.client[0].run_command(ldb_cmd2, raiseonerr=False)
        cmd1_output = multihost.client[0].get_file_contents('/tmp/file_ldb1').decode('utf-8')
        cmd2_output = multihost.client[0].get_file_contents('/tmp/file_ldb2').decode('utf-8')
        assert cmd.returncode == 0, f'{invalidate_cache} did not execute successfully'
        assert cmd1.returncode == 0, f'{ldb_cmd1} did not execute successfully'
        assert "dataExpireTimestamp: 1\n" in cmd1_output, "dataExpireTimestamp not found in /tmp/file_ldb1"
        assert cmd2.returncode == 0, f'{ldb_cmd2} did not execute successfully'
        assert "dataExpireTimestamp: 1\n" in cmd2_output, "dataExpireTimestamp not found in /tmp/file_ldb2"

    @staticmethod
    def test_0011_Refresh_group_record_after_expiry(multihost):
        """
        :title: IDM-SSSD-TC: ldap_provider: cache_performance: Refresh group record after expiry and \
            verify the cache updates
        :id: d1e735df-cd09-43cf-b069-48c244acaece
        :steps:
          1. Execute getent to fetch group details.
          2. Execute the ldbsearch command and check for dataExpireTimestamp in the cache db output
          3. Execute the ldbsearch command and check for dataExpireTimestamp in the timestamps db output
        :expectedresults:
          1. Group details should be successfully fetched.
          2. dataExpireTimestamp should be present in cache db output.
          3. dataExpireTimestamp should not be present in timestamps db output.
        """
        client = sssdTools(multihost.client[0])
        client.get_getent_group('ldapusers')
        ldb_cmd = f"ldbsearch -H /var/lib/sss/db/cache_{ds_instance_name}.ldb \
                    -b 'name=ldapusers@{ds_instance_name},cn=groups,cn={ds_instance_name},cn=sysdb' > /tmp/file_ldb"
        ldb_cmd2 = f"ldbsearch -H /var/lib/sss/db/timestamps_{ds_instance_name}.ldb \
                    -b 'name=ldapusers@{ds_instance_name},cn=groups,cn={ds_instance_name},cn=sysdb' > /tmp/file_ldb2"
        cmd = multihost.client[0].run_command(ldb_cmd, raiseonerr=False)
        cmd2 = multihost.client[0].run_command(ldb_cmd2, raiseonerr=False)
        cmd_output = multihost.client[0].get_file_contents('/tmp/file_ldb').decode('utf-8')
        cmd2_output = multihost.client[0].get_file_contents('/tmp/file_ldb2').decode('utf-8')
        assert cmd.returncode == 0, f'{ldb_cmd} did not execute successfully'
        assert "dataExpireTimestamp: 1\n" in cmd_output, "dataExpireTimestamp not found in /tmp/file_ldb"
        assert cmd2.returncode == 0, f'{ldb_cmd2} did not execute successfully'
        assert "dataExpireTimestamp: 1\n" not in cmd2_output, "dataExpireTimestamp found in /tmp/file_ldb2"

    @staticmethod
    def test_0012_Set_refresh_expired_interval_to_40(multihost, backupsssdconf):
        """
        :title: IDM-SSSD-TC: ldap_provider: cache_performance: Set refresh expired interval to 40 and \
            verify Group record updates
        :id: 1905e2ae-3e1a-40e4-8aff-378de9d8f45b
        :setup:
          1. Set the refresh_expired_interval in the sssd.conf
        :steps:
          1. Execute getent to fetch group details.
          2. Invalidate the existing cache
          3. Execute the ldbsearch command and check for dataExpireTimestamp in the cache db output
          4. Execute the ldbsearch command and check for dataExpireTimestamp in the timestamps db output
        :expectedresults:
          1. Group details should be successfully fetched.
          2. Cache sucessfully get invalidated.
          3. dataExpireTimestamp should be present in cache db output.
          4. dataExpireTimestamp should not be present in timestamps db output.
        """
        client = sssdTools(multihost.client[0])
        domain_section = f'domain/{ds_instance_name}'
        domain_params = {'entry_cache_timeout': 160,
                         'refresh_expired_interval': 40
                         }
        client.sssd_conf(domain_section, domain_params)
        client.clear_sssd_cache()
        client.get_getent_group('ldapusers')

        invalidate_cache = "sss_cache -E"
        cmd = multihost.client[0].run_command(invalidate_cache, raiseonerr=False)
        time.sleep(45)
        ldb_cmd = f"ldbsearch -H /var/lib/sss/db/cache_{ds_instance_name}.ldb \
                    -b 'name=ldapusers@{ds_instance_name},cn=groups,cn={ds_instance_name},cn=sysdb' > /tmp/file_ldb"
        ldb_cmd2 = f"ldbsearch -H /var/lib/sss/db/timestamps_{ds_instance_name}.ldb \
                    -b 'name=ldapusers@{ds_instance_name},cn=groups,cn={ds_instance_name},cn=sysdb' > /tmp/file_ldb2"
        cmd1 = multihost.client[0].run_command(ldb_cmd, raiseonerr=False)
        cmd2 = multihost.client[0].run_command(ldb_cmd2, raiseonerr=False)
        cmd1_output = multihost.client[0].get_file_contents('/tmp/file_ldb').decode('utf-8')
        cmd2_output = multihost.client[0].get_file_contents('/tmp/file_ldb2').decode('utf-8')
        assert cmd.returncode == 0, f'{invalidate_cache} did not execute successfully'
        assert cmd1.returncode == 0, f'{ldb_cmd} did not execute successfully'
        assert "dataExpireTimestamp: 1\n" in cmd1_output, "dataExpireTimestamp not found in /tmp/file_ldb"
        assert cmd2.returncode == 0, f'{ldb_cmd2} did not execute successfully'
        assert "dataExpireTimestamp: 1\n" not in cmd2_output, "dataExpireTimestamp found in /tmp/file_ldb2"

    @pytest.mark.converted('test_sss_cache.py', 'test_sss_cache__check_ldb_updates_when_user_is_deleted_and_modified')
    @staticmethod
    def test_0013_Modify_User_Attribute(multihost):
        """
        :title: IDM-SSSD-TC: ldap_provider: cache_performance: Modify user attribute and verify cache updates
        :id: ae02729b-8ac9-491f-8a0d-9902ee4948fa
        :steps:
          1. Execute getent to fetch group details.
          2. Invalidate the existing cache
          3. Execute the ldapmodify command to modify user's loginShell details
          4. Execute the ldbsearch command and check for loginShell in the cache db output
          5. Execute the ldbsearch command and check for dataExpireTimestamp in the timestamps db output
        :expectedresults:
          1. Group details should be successfully fetched.
          2. Cache sucessfully get invalidated.
          3. Ldapmodify command executes successfully.
          4. Updated loginShell value should be present in cache db output.
          5. dataExpireTimestamp should not be present in timestamps db output.
        """
        client = sssdTools(multihost.client[0])
        client.clear_sssd_cache()
        client.get_getent_passwd('foo1')

        invalidate_cache = "sss_cache -E"
        cmd = multihost.client[0].run_command(invalidate_cache, raiseonerr=False)

        user = 'uid=foo1,ou=People,dc=example,dc=test'
        content = f'''dn: {user}\nchangetype: modify\nreplace: loginShell\nloginShell: /bin/sh\n'''
        multihost.client[0].put_file_contents("/tmp/changes.ldif", content)
        ldap_cmd = f'ldapmodify -x -H ldap://{multihost.master[0].sys_hostname}' \
                   f' -D "cn=Directory Manager" -w "Secret123" -f /tmp/changes.ldif'
        cmd2 = multihost.client[0].run_command(ldap_cmd, raiseonerr=False)
        client.get_getent_passwd('foo1')

        ldb_cmd = f"ldbsearch -H /var/lib/sss/db/cache_{ds_instance_name}.ldb \
                    -b 'name=foo1@{ds_instance_name},cn=users,cn={ds_instance_name},cn=sysdb' > /tmp/file_ldb"
        ldb_cmd2 = f"ldbsearch -H /var/lib/sss/db/timestamps_{ds_instance_name}.ldb \
                    -b 'name=foo1@{ds_instance_name},cn=users,cn={ds_instance_name},cn=sysdb' > /tmp/file_ldb2"
        cmd3 = multihost.client[0].run_command(ldb_cmd, raiseonerr=False)
        cmd4 = multihost.client[0].run_command(ldb_cmd2, raiseonerr=False)
        cmd3_output = multihost.client[0].get_file_contents('/tmp/file_ldb').decode('utf-8')
        cmd4_output = multihost.client[0].get_file_contents('/tmp/file_ldb2').decode('utf-8')
        assert cmd.returncode == 0, f'{invalidate_cache} did not execute successfully'
        assert cmd2.returncode == 0, f'{ldap_cmd} did not execute successfully'
        assert cmd3.returncode == 0, f'{ldb_cmd} did not execute successfully'
        assert "/bin/sh" in cmd3_output, "Updated loginShell value not found in /tmp/file_ldb"
        assert cmd4.returncode == 0, f'{ldb_cmd2} did not execute successfully'
        assert "dataExpireTimestamp: 1\n" not in cmd4_output, "dataExpireTimestamp found in /tmp/file_ldb2"

    @pytest.mark.converted('test_sss_cache.py', 'test_sss_cache__check_ldb_updates_when_user_is_deleted_and_modified')
    @staticmethod
    def test_0014_Delete_an_existing_user(multihost):
        """
        :title: IDM-SSSD-TC: ldap_provider: cache_performance: Delete an existing user and verify cache updates
        :id: 2c05ee2c-d5d6-4c98-9b1d-e349df5d7780
        :steps:
          1. Execute getent to fetch user details.
          2. Invalidate the existing cache
          3. Execute the ldapdelete command to delete user foo1.
          4. Execute the ldbsearch command and check if user foo1 is present in the cache db output
          5. Execute the ldbsearch command and check if user foo1 is present in the timestamps db output
        :expectedresults:
          1. User details should be successfully fetched.
          2. Cache sucessfully get invalidated.
          3. Ldapdelete command executes successfully.
          4. User foo1 should not be present in cache db output.
          5. User foo1 should not be present in timestamps db output.
        """
        client = sssdTools(multihost.client[0])
        client.clear_sssd_cache()
        client.get_getent_passwd('foo1')

        invalidate_cache = "sss_cache -E"
        cmd = multihost.client[0].run_command(invalidate_cache, raiseonerr=False)

        ldap_cmd = f'ldapdelete -x -H ldap://{multihost.master[0].sys_hostname} \
                   -D "cn=Directory Manager" -w "Secret123" uid=foo1,ou=People,dc=example,dc=test'

        delete_cmd = multihost.client[0].run_command(ldap_cmd, raiseonerr=False)
        client.get_getent_passwd('foo1')
        ldb_cmd = f"ldbsearch -H /var/lib/sss/db/cache_{ds_instance_name}.ldb \
                    -b 'name=foo1@{ds_instance_name},cn=users,cn={ds_instance_name},cn=sysdb' > /tmp/file_ldb"
        ldb_cmd2 = f"ldbsearch -H /var/lib/sss/db/timestamps_{ds_instance_name}.ldb \
                    -b 'name=foo1@{ds_instance_name},cn=users,cn={ds_instance_name},cn=sysdb' > /tmp/file_ldb2"
        cmd1 = multihost.client[0].run_command(ldb_cmd, raiseonerr=False)
        cmd2 = multihost.client[0].run_command(ldb_cmd2, raiseonerr=False)
        cmd1_output = multihost.client[0].get_file_contents('/tmp/file_ldb').decode('utf-8')
        cmd2_output = multihost.client[0].get_file_contents('/tmp/file_ldb2').decode('utf-8')
        assert cmd.returncode == 0, f'{invalidate_cache} did not execute successfully'
        assert delete_cmd.returncode == 0, f'{ldap_cmd} did not execute successfully'
        assert cmd1.returncode == 0, f'{ldb_cmd} did not execute successfully'
        assert "foo1@example1" not in cmd1_output, "User foo1 found in /tmp/file_ldb"
        assert cmd2.returncode == 0, f'{ldb_cmd2} did not execute successfully'
        assert "foo1@example1" not in cmd2_output, "User foo1 found in /tmp/file_ldb2"



import sys
import gettext

PACKAGE = 'sss_daemon'
LOCALEDIR = '/usr/share/locale'

translation = gettext.translation(PACKAGE, LOCALEDIR, fallback=True)
if sys.version_info[0] > 2:
    _ = translation.gettext
else:
    _ = translation.ugettext


class SSSDOptions(object):
    def __init__(self):
        pass

    option_strings = {
        # [service]
        'debug': _('Set the verbosity of the debug logging'),
        'debug_level': _('Set the verbosity of the debug logging'),
        'debug_timestamps': _('Include timestamps in debug logs'),
        'debug_microseconds': _('Include microseconds in timestamps in debug logs'),
        'debug_backtrace_enabled': _('Enable/disable debug backtrace'),
        'timeout': _('Watchdog timeout before restarting service'),
        'command': _('Command to start service'),
        'fd_limit': _('The number of file descriptors that may be opened by this responder'),
        'client_idle_timeout': _('Idle time before automatic disconnection of a client'),
        'responder_idle_timeout': _('Idle time before automatic shutdown of the responder'),
        'cache_first': _('Always query all the caches before querying the Data Providers'),
        'offline_timeout': _('When SSSD switches to offline mode the amount of time before it tries to go back online '
                             'will increase based upon the time spent disconnected. This value is in seconds and '
                             'calculated by the following: offline_timeout + random_offset.'),

        # [sssd]
        'services': _('SSSD Services to start'),
        'domains': _('SSSD Domains to start'),
        're_expression': _('Regex to parse username and domain'),
        'full_name_format': _('Printf-compatible format for displaying fully-qualified names'),
        'krb5_rcache_dir': _('Directory on the filesystem where SSSD should store Kerberos replay cache files.'),
        'default_domain_suffix': _('Domain to add to names without a domain component.'),
        'user': _('The user to drop privileges to'),
        'certificate_verification': _('Tune certificate verification'),
        'override_space': _('All spaces in group or user names will be replaced with this character'),
        'disable_netlink': _('Tune sssd to honor or ignore netlink state changes'),
        'enable_files_domain': _('Enable or disable the implicit files domain'),
        'domain_resolution_order': _('A specific order of the domains to be looked up'),
        'monitor_resolv_conf': _('Controls if SSSD should monitor the state of resolv.conf to identify when it needs '
                                 'to update its internal DNS resolver.'),
        'try_inotify': _('SSSD monitors the state of resolv.conf to identify when it needs to update its internal DNS '
                         'resolver. By default, we will attempt to use inotify for this, and will fall back to '
                         'polling resolv.conf every five seconds if inotify cannot be used.'),
        'implicit_pac_responder': _('Run PAC responder automatically for AD and IPA provider'),
        'core_dumpable': _('Enable or disable core dumps for all SSSD processes.'),
        'passkey_verification': _('Tune passkey verification behavior'),

        # [nss]
        'enum_cache_timeout': _('Enumeration cache timeout length (seconds)'),
        'entry_cache_no_wait_timeout': _('Entry cache background update timeout length (seconds)'),
        'entry_negative_timeout': _('Negative cache timeout length (seconds)'),
        'local_negative_timeout': _('Files negative cache timeout length (seconds)'),
        'filter_users': _('Users that SSSD should explicitly ignore'),
        'filter_groups': _('Groups that SSSD should explicitly ignore'),
        'filter_users_in_groups': _('Should filtered users appear in groups'),
        'pwfield': _('The value of the password field the NSS provider should return'),
        'override_homedir': _('Override homedir value from the identity provider with this value'),
        'fallback_homedir': _('Substitute empty homedir value from the identity provider with this value'),
        'override_shell': _('Override shell value from the identity provider with this value'),
        'allowed_shells': _('The list of shells users are allowed to log in with'),
        'vetoed_shells': _('The list of shells that will be vetoed, and replaced with the fallback shell'),
        'shell_fallback': _('If a shell stored in central directory is allowed but not available, use this fallback'),
        'default_shell': _('Shell to use if the provider does not list one'),
        'memcache_timeout': _('How long will be in-memory cache records valid'),
        'memcache_size_passwd': _(
            'Size (in megabytes) of the data table allocated inside fast in-memory cache for passwd requests'),
        'memcache_size_group': _(
            'Size (in megabytes) of the data table allocated inside fast in-memory cache for group requests'),
        'memcache_size_initgroups': _(
            'Size (in megabytes) of the data table allocated inside fast in-memory cache for initgroups requests'),
        'homedir_substring': _('The value of this option will be used in the expansion of the override_homedir option '
                               'if the template contains the format string %H.'),
        'get_domains_timeout': _('Specifies time in seconds for which the list of subdomains will be considered '
                                 'valid.'),
        'entry_cache_nowait_percentage': _('The entry cache can be set to automatically update entries in the '
                                           'background if they are requested beyond a percentage of the '
                                           'entry_cache_timeout value for the domain.'),

        # [pam]
        'offline_credentials_expiration': _('How long to allow cached logins between online logins (days)'),
        'offline_failed_login_attempts': _('How many failed logins attempts are allowed when offline'),
        'offline_failed_login_delay': _(
            'How long (minutes) to deny login after offline_failed_login_attempts has been reached'),
        'pam_verbosity': _('What kind of messages are displayed to the user during authentication'),
        'pam_response_filter': _('Filter PAM responses sent to the pam_sss'),
        'pam_id_timeout': _('How many seconds to keep identity information cached for PAM requests'),
        'pam_pwd_expiration_warning': _('How many days before password expiration a warning should be displayed'),
        'pam_trusted_users': _('List of trusted uids or user\'s name'),
        'pam_public_domains': _('List of domains accessible even for untrusted users.'),
        'pam_account_expired_message': _('Message printed when user account is expired.'),
        'pam_account_locked_message': _('Message printed when user account is locked.'),
        'pam_cert_auth': _('Allow certificate based/Smartcard authentication.'),
        'pam_cert_db_path': _('Path to certificate database with PKCS#11 modules.'),
        'pam_cert_verification': _('Tune certificate verification for PAM authentication.'),
        'p11_child_timeout': _('How many seconds will pam_sss wait for p11_child to finish'),
        'pam_app_services': _('Which PAM services are permitted to contact application domains'),
        'pam_p11_allowed_services': _('Allowed services for using smartcards'),
        'p11_wait_for_card_timeout': _('Additional timeout to wait for a card if requested'),
        'p11_uri': _('PKCS#11 URI to restrict the selection of devices for Smartcard authentication'),
        'pam_initgroups_scheme': _('When shall the PAM responder force an initgroups request'),
        'pam_gssapi_services': _('List of PAM services that are allowed to authenticate with GSSAPI.'),
        'pam_gssapi_check_upn': _('Whether to match authenticated UPN with target user'),
        'pam_gssapi_indicators_map': _('List of pairs <PAM service>:<authentication indicator> that '
                                       'must be enforced for PAM access with GSSAPI authentication'),
        'pam_passkey_auth': _('Allow passkey device authentication.'),
        'passkey_child_timeout': _('How many seconds will pam_sss wait for passkey_child to finish'),
        'passkey_debug_libfido2': _('Enable debugging in the libfido2 library'),

        # [sudo]
        'sudo_timed': _('Whether to evaluate the time-based attributes in sudo rules'),
        'sudo_inverse_order': _('If true, SSSD will switch back to lower-wins ordering logic'),
        'sudo_threshold': _('Maximum number of rules that can be refreshed at once. If this is exceeded, full refresh '
                            'is performed.'),

        # [autofs]
        'autofs_negative_timeout': _('Negative cache timeout length (seconds)'),

        # [ssh]
        'ssh_hash_known_hosts': _('Whether to hash host names and addresses in the known_hosts file'),
        'ssh_known_hosts_timeout': _('How many seconds to keep a host in the known_hosts file after its host keys '
                                     'were requested'),
        'ca_db': _('Path to storage of trusted CA certificates'),
        'ssh_use_certificate_keys': _('Allow to generate ssh-keys from certificates'),
        'ssh_use_certificate_matching_rules': _('Use the following matching rules to filter the certificates for '
                                                'ssh-key generation'),

        # [pac]
        'allowed_uids': _('List of UIDs or user names allowed to access the PAC responder'),
        'pac_lifetime': _('How long the PAC data is considered valid'),
        'pac_check': _('Validate the PAC'),

        # [ifp]
        'user_attributes': _('List of user attributes the InfoPipe is allowed to publish'),

        # [session_recording]
        'scope': _('One of the following strings specifying the scope of session recording: none - No users are '
                   'recorded. some - Users/groups specified by users and groups options are recorded. all - All users '
                   'are recorded.'),
        'users': _('A comma-separated list of users which should have session recording enabled. Matches user names '
                   'as returned by NSS. I.e. after the possible space replacement, case changes, etc.'),
        'groups': _('A comma-separated list of groups, members of which should have session recording enabled. '
                    'Matches group names as returned by NSS. I.e. after the possible space replacement, case changes, '
                    'etc.'),
        'exclude_users': _('A comma-separated list of users to be excluded from recording, only when scope=all'),
        'exclude_groups': _('A comma-separated list of groups, members of which should be excluded from recording, '
                            ' only when scope=all. '),

        # [provider]
        'id_provider': _('Identity provider'),
        'auth_provider': _('Authentication provider'),
        'access_provider': _('Access control provider'),
        'chpass_provider': _('Password change provider'),
        'sudo_provider': _('SUDO provider'),
        'autofs_provider': _('Autofs provider'),
        'hostid_provider': _('Host identity provider'),
        'selinux_provider': _('SELinux provider'),
        'session_provider': _('Session management provider'),
        'resolver_provider': _('Resolver provider'),

        # [domain]
        'domain_type': _('Whether the domain is usable by the OS or by applications'),
        'enabled': _('Enable or disable the domain'),
        'min_id': _('Minimum user ID'),
        'max_id': _('Maximum user ID'),
        'enumerate': _('Enable enumerating all users/groups'),
        'cache_credentials': _('Cache credentials for offline login'),
        'use_fully_qualified_names': _('Display users/groups in fully-qualified form'),
        'ignore_group_members': _('Don\'t include group members in group lookups'),
        'entry_cache_timeout': _('Entry cache timeout length (seconds)'),
        'lookup_family_order': _('Restrict or prefer a specific address family when performing DNS lookups'),
        'account_cache_expiration': _('How long to keep cached entries after last successful login (days)'),
        'dns_resolver_server_timeout': _('How long should SSSD talk to single DNS server before trying next server ('
                                         'miliseconds)'),
        'dns_resolver_op_timeout': _('How long should keep trying to resolve single DNS query (seconds)'),
        'dns_resolver_timeout': _('How long to wait for replies from DNS when resolving servers (seconds)'),
        'dns_discovery_domain': _('The domain part of service discovery DNS query'),
        'failover_primary_timeout': _('Specifies the interval, in seconds, that SSSD waits before attempting to reconnect to the primary '
                                      'server after a successful connection to the backup server'),
        'override_gid': _('Override GID value from the identity provider with this value'),
        'case_sensitive': _('Treat usernames as case sensitive'),
        'entry_cache_user_timeout': _('Entry cache timeout length (seconds)'),
        'entry_cache_group_timeout': _('Entry cache timeout length (seconds)'),
        'entry_cache_netgroup_timeout': _('Entry cache timeout length (seconds)'),
        'entry_cache_service_timeout': _('Entry cache timeout length (seconds)'),
        'entry_cache_autofs_timeout': _('Entry cache timeout length (seconds)'),
        'entry_cache_sudo_timeout': _('Entry cache timeout length (seconds)'),
        'entry_cache_resolver_timeout': _('Entry cache timeout length (seconds)'),
        'refresh_expired_interval': _('How often should expired entries be refreshed in background'),
        'refresh_expired_interval_offset': _("Maximum period deviation when refreshing expired entries in background"),
        'dyndns_update': _("Whether to automatically update the client's DNS entry"),
        'dyndns_update_per_family': _('Whether DNS update of A and AAAA record should be performed '
                                      'in one update or in two separate updates'),
        'dyndns_ttl': _("The TTL to apply to the client's DNS entry after updating it"),
        'dyndns_iface': _("The interface whose IP should be used for dynamic DNS updates"),
        'dyndns_refresh_interval': _("How often to periodically update the client's DNS entry"),
        'dyndns_refresh_interval_offset': _("Maximum period deviation when updating the client's DNS entry"),
        'dyndns_update_ptr': _("Whether the provider should explicitly update the PTR record as well"),
        'dyndns_force_tcp': _("Whether the nsupdate utility should default to using TCP"),
        'dyndns_auth': _("What kind of authentication should be used to perform the DNS update"),
        'dyndns_server': _("Override the DNS server used to perform the DNS update"),
        'dyndns_dot_cacert': _("The file of the certificate authorities certificates for DoT"),
        'dyndns_dot_cert': _("The certificate(s) file for authentication for the DoT transport"),
        'dyndns_dot_key': _("The key file for authenticated encryption for the DoT transport"),
        'subdomain_enumerate': _('Control enumeration of trusted domains'),
        'subdomain_refresh_interval': _('How often should subdomains list be refreshed'),
        'subdomain_refresh_interval_offset': _('Maximum period deviation when refreshing the subdomain list'),
        'subdomain_inherit': _('List of options that should be inherited into a subdomain'),
        'subdomain_homedir': _('Default subdomain homedir value'),
        'cached_auth_timeout': _('How long can cached credentials be used for cached authentication'),
        'auto_private_groups': _('Whether to automatically create private groups for users'),
        'pwd_expiration_warning': _('Display a warning N days before the password expires.'),
        'realmd_tags': _('Various tags stored by the realmd configuration service for this domain.'),
        'subdomains_provider': _('The provider which should handle fetching of subdomains. This value should be '
                                 'always the same as id_provider.'),
        'entry_cache_ssh_host_timeout': _('How many seconds to keep a host ssh key after refresh. IE how long to '
                                          'cache the host key for.'),
        'cache_credentials_minimal_first_factor_length': _('If 2-Factor-Authentication (2FA) is used and credentials '
                                                           'should be saved this value determines the minimal length '
                                                           'the first authentication factor (long term password) must '
                                                           'have to be saved as SHA512 hash into the cache.'),
        'local_auth_policy': _('Local authentication methods policy '),

        # [provider/ipa]
        'ipa_domain': _('IPA domain'),
        'ipa_server': _('IPA server address'),
        'ipa_backup_server': _('Address of backup IPA server'),
        'ipa_hostname': _('IPA client hostname'),
        'ipa_dyndns_update': _("Whether to automatically update the client's DNS entry in FreeIPA"),
        'ipa_dyndns_ttl': _("The TTL to apply to the client's DNS entry after updating it"),
        'ipa_dyndns_iface': _("The interface whose IP should be used for dynamic DNS updates"),
        'ipa_hbac_search_base': _("Search base for HBAC related objects"),
        'ipa_hbac_refresh': _("The amount of time between lookups of the HBAC rules against the IPA server"),
        'ipa_selinux_refresh': _("The amount of time in seconds between lookups of the SELinux maps against the IPA "
                                 "server"),
        'ipa_hbac_support_srchost': _("If set to false, host argument given by PAM will be ignored"),
        'ipa_automount_location': _("The automounter location this IPA client is using"),
        'ipa_master_domain_search_base': _("Search base for object containing info about IPA domain"),
        'ipa_ranges_search_base': _("Search base for objects containing info about ID ranges"),
        'ipa_enable_dns_sites': _("Enable DNS sites - location based service discovery"),
        'ipa_views_search_base': _("Search base for view containers"),
        'ipa_view_class': _("Objectclass for view containers"),
        'ipa_view_name': _("Attribute with the name of the view"),
        'ipa_override_object_class': _("Objectclass for override objects"),
        'ipa_anchor_uuid': _("Attribute with the reference to the original object"),
        'ipa_user_override_object_class': _("Objectclass for user override objects"),
        'ipa_group_override_object_class': _("Objectclass for group override objects"),
        'ipa_deskprofile_search_base': _("Search base for Desktop Profile related objects"),
        'ipa_deskprofile_refresh': _("The amount of time in seconds between lookups of the Desktop Profile rules "
                                     "against the IPA server"),
        'ipa_deskprofile_request_interval': _("The amount of time in minutes between lookups of Desktop Profiles "
                                              "rules against the IPA server when the last request did not find any "
                                              "rule"),
        'ipa_subid_ranges_search_base': _("Search base for SUBID ranges"),
        'ipa_access_order': _("Which rules should be used to evaluate access control"),
        'ipa_host_fqdn': _('The LDAP attribute that contains FQDN of the host.'),
        'ipa_host_object_class': _('The object class of a host entry in LDAP.'),
        'ipa_host_search_base': _('Use the given string as search base for host objects.'),
        'ipa_host_ssh_public_key': _('The LDAP attribute that contains the host\'s SSH public keys.'),
        'ipa_netgroup_domain': _('The LDAP attribute that contains NIS domain name of the netgroup.'),
        'ipa_netgroup_member': _('The LDAP attribute that contains the names of the netgroup\'s members.'),
        'ipa_netgroup_member_ext_host': _('The LDAP attribute that lists FQDNs of hosts and host groups that are '
                                          'members of the netgroup.'),
        'ipa_netgroup_member_host': _('The LDAP attribute that lists hosts and host groups that are direct members of '
                                      'the netgroup.'),
        'ipa_netgroup_member_of': _('The LDAP attribute that lists netgroup\'s memberships.'),
        'ipa_netgroup_member_user': _('The LDAP attribute that lists system users and groups that are direct members '
                                      'of the netgroup.'),
        'ipa_netgroup_name': _('The LDAP attribute that corresponds to the netgroup name.'),
        'ipa_netgroup_object_class': _('The object class of a netgroup entry in LDAP.'),
        'ipa_netgroup_uuid': _('The LDAP attribute that contains the UUID/GUID of an LDAP netgroup object.'),
        'ipa_selinux_usermap_enabled': _('The LDAP attribute that contains whether or not is user map enabled for '
                                         'usage.'),
        'ipa_selinux_usermap_host_category': _('The LDAP attribute that contains host category such as \'all\'.'),
        'ipa_selinux_usermap_member_host': _('The LDAP attribute that contains all hosts / hostgroups this rule match '
                                             'against.'),
        'ipa_selinux_usermap_member_user': _('The LDAP attribute that contains all users / groups this rule match '
                                             'against.'),
        'ipa_selinux_usermap_name': _('The LDAP attribute that contains the name of SELinux usermap.'),
        'ipa_selinux_usermap_object_class': _('The object class of a host entry in LDAP.'),
        'ipa_selinux_usermap_see_also': _('The LDAP attribute that contains DN of HBAC rule which can be used for '
                                          'matching instead of memberUser and memberHost.'),
        'ipa_selinux_usermap_selinux_user': _('The LDAP attribute that contains SELinux user string itself.'),
        'ipa_selinux_usermap_user_category': _('The LDAP attribute that contains user category such as \'all\'.'),
        'ipa_selinux_usermap_uuid': _('The LDAP attribute that contains unique ID of the user map.'),
        'ipa_server_mode': _('The option denotes that the SSSD is running on IPA server and should perform lookups of '
                             'users and groups from trusted domains differently.'),
        'ipa_subdomains_search_base': _('Use the given string as search base for trusted domains.'),

        # [provider/ad]
        'ad_domain': _('Active Directory domain'),
        'ad_enabled_domains': _('Enabled Active Directory domains'),
        'ad_server': _('Active Directory server address'),
        'ad_backup_server': _('Active Directory backup server address'),
        'ad_hostname': _('Active Directory client hostname'),
        'ad_enable_dns_sites': _('Enable DNS sites - location based service discovery'),
        'ad_access_filter': _('LDAP filter to determine access privileges'),
        'ad_enable_gc': _('Whether to use the Global Catalog for lookups'),
        'ad_gpo_access_control': _('Operation mode for GPO-based access control'),
        'ad_gpo_cache_timeout': _("The amount of time between lookups of the GPO policy files against the AD server"),
        'ad_gpo_map_interactive': _('PAM service names that map to the GPO (Deny)InteractiveLogonRight '
                                    'policy settings'),
        'ad_gpo_map_remote_interactive': _('PAM service names that map to the GPO (Deny)RemoteInteractiveLogonRight '
                                           'policy settings'),
        'ad_gpo_map_network': _('PAM service names that map to the GPO (Deny)NetworkLogonRight policy settings'),
        'ad_gpo_map_batch': _('PAM service names that map to the GPO (Deny)BatchLogonRight policy settings'),
        'ad_gpo_map_service': _('PAM service names that map to the GPO (Deny)ServiceLogonRight policy settings'),
        'ad_gpo_map_permit': _('PAM service names for which GPO-based access is always granted'),
        'ad_gpo_map_deny': _('PAM service names for which GPO-based access is always denied'),
        'ad_gpo_default_right': _('Default logon right (or permit/deny) to use for unmapped PAM service names'),
        'ad_site': _('a particular site to be used by the client'),
        'ad_maximum_machine_account_password_age': _('Maximum age in days before the machine account password should '
                                                     'be renewed'),
        'ad_machine_account_password_renewal_opts': _('Option for tuning the machine account renewal task'),
        'ad_update_samba_machine_account_password': _('Whether to update the machine account password in the Samba '
                                                      'database'),
        'ad_use_ldaps': _('Use LDAPS port for LDAP and Global Catalog requests'),
        'ad_allow_remote_domain_local_groups': _('Do not filter domain local groups from other domains'),

        # [provider/krb5]
        'krb5_kdcip': _('Kerberos server address'),
        'krb5_server': _('Kerberos server address'),
        'krb5_backup_server': _('Kerberos backup server address'),
        'krb5_realm': _('Kerberos realm'),
        'krb5_auth_timeout': _('Authentication timeout'),
        'krb5_use_kdcinfo': _('Whether to create kdcinfo files'),
        'krb5_confd_path': _('Where to drop krb5 config snippets'),

        # [provider/krb5/auth]
        'krb5_ccachedir': _('Directory to store credential caches'),
        'krb5_ccname_template': _("Location of the user's credential cache"),
        'krb5_keytab': _("Location of the keytab to validate credentials"),
        'krb5_validate': _("Enable credential validation"),
        'krb5_store_password_if_offline': _("Store password if offline for later online authentication"),
        'krb5_renewable_lifetime': _("Renewable lifetime of the TGT"),
        'krb5_lifetime': _("Lifetime of the TGT"),
        'krb5_renew_interval': _("Time between two checks for renewal"),
        'krb5_use_fast': _("Enables FAST"),
        'krb5_fast_principal': _("Selects the principal to use for FAST"),
        'krb5_fast_use_anonymous_pkinit': _("Use anonymous PKINIT to request FAST credentials"),
        'krb5_canonicalize': _("Enables principal canonicalization"),
        'krb5_use_enterprise_principal': _("Enables enterprise principals"),
        'krb5_use_subdomain_realm': _("Enables using of subdomains realms for authentication"),
        'krb5_map_user': _('A mapping from user names to Kerberos principal names'),

        # [provider/krb5/chpass]
        'krb5_kpasswd': _('Server where the change password service is running if not on the KDC'),
        'krb5_backup_kpasswd': _('Server where the change password service is running if not on the KDC'),

        # [provider/ldap]
        'ldap_uri': _('ldap_uri, The URI of the LDAP server'),
        'ldap_backup_uri': _('ldap_backup_uri, The URI of the LDAP server'),
        'ldap_search_base': _('The default base DN'),
        'ldap_schema': _('The Schema Type in use on the LDAP server, rfc2307'),
        'ldap_pwmodify_mode': _('Mode used to change user password'),
        'ldap_default_bind_dn': _('The default bind DN'),
        'ldap_default_authtok_type': _('The type of the authentication token of the default bind DN'),
        'ldap_default_authtok': _('The authentication token of the default bind DN'),
        'ldap_network_timeout': _('Length of time to attempt connection'),
        'ldap_opt_timeout': _('Length of time to attempt synchronous LDAP operations'),
        'ldap_offline_timeout': _('Length of time between attempts to reconnect while offline'),
        'ldap_force_upper_case_realm': _('Use only the upper case for realm names'),
        'ldap_tls_cacert': _('File that contains CA certificates'),
        'ldap_tls_cacertdir': _('Path to CA certificate directory'),
        'ldap_tls_cert': _('File that contains the client certificate'),
        'ldap_tls_key': _('File that contains the client key'),
        'ldap_tls_cipher_suite': _('List of possible ciphers suites'),
        'ldap_tls_reqcert': _('Require TLS certificate verification'),
        'ldap_sasl_mech': _('Specify the sasl mechanism to use'),
        'ldap_sasl_authid': _('Specify the sasl authorization id to use'),
        'ldap_sasl_realm': _('Specify the sasl authorization realm to use'),
        'ldap_sasl_minssf': _('Specify the minimal SSF for LDAP sasl authorization'),
        'ldap_sasl_maxssf': _('Specify the maximal SSF for LDAP sasl authorization'),
        'ldap_krb5_keytab': _('Kerberos service keytab'),
        'ldap_krb5_init_creds': _('Use Kerberos auth for LDAP connection'),
        'ldap_referrals': _('Follow LDAP referrals'),
        'ldap_krb5_ticket_lifetime': _('Lifetime of TGT for LDAP connection'),
        'ldap_deref': _('How to dereference aliases'),
        'ldap_dns_service_name': _('Service name for DNS service lookups'),
        'ldap_page_size': _('The number of records to retrieve in a single LDAP query'),
        'ldap_deref_threshold': _('The number of members that must be missing to trigger a full deref'),
        'ldap_ignore_unreadable_references': _('Ignore unreadable LDAP references'),
        'ldap_sasl_canonicalize': _('Whether the LDAP library should perform a reverse lookup to canonicalize the '
                                    'host name during a SASL bind'),
        'ldap_rfc2307_fallback_to_local_users': _('Allows to retain local users as members of an LDAP group for '
                                                  'servers that use the RFC2307 schema.'),

        'ldap_entry_usn': _('entryUSN attribute'),
        'ldap_rootdse_last_usn': _('lastUSN attribute'),

        'ldap_connection_expiration_timeout': _('How long to retain a connection to the LDAP server before '
                                                'disconnecting'),

        'ldap_disable_paging': _('Disable the LDAP paging control'),
        'ldap_disable_range_retrieval': _('Disable Active Directory range retrieval'),
        'ldap_use_ppolicy': _('Use the ppolicy extension'),
        'ldap_ppolicy_pwd_change_threshold': _('Force a password change when remaining grace logins reach or go below this threshold'),

        # [provider/ldap/id]
        'ldap_search_timeout': _('Length of time to wait for a search request'),
        'ldap_enumeration_search_timeout': _('Length of time to wait for a enumeration request'),
        'ldap_enumeration_refresh_timeout': _('Length of time between enumeration updates'),
        'ldap_enumeration_refresh_offset': _('Maximum period deviation between enumeration updates'),
        'ldap_purge_cache_timeout': _('Length of time between cache cleanups'),
        'ldap_purge_cache_offset': _('Maximum time deviation between cache cleanups'),
        'ldap_id_use_start_tls': _('Require TLS for ID lookups'),
        'ldap_id_mapping': _('Use ID-mapping of objectSID instead of pre-set IDs'),
        'ldap_user_search_base': _('Base DN for user lookups'),
        'ldap_user_search_scope': _('Scope of user lookups'),
        'ldap_user_search_filter': _('Filter for user lookups'),
        'ldap_user_object_class': _('Objectclass for users'),
        'ldap_user_name': _('Username attribute'),
        'ldap_user_uid_number': _('UID attribute'),
        'ldap_user_gid_number': _('Primary GID attribute'),
        'ldap_user_gecos': _('GECOS attribute'),
        'ldap_user_home_directory': _('Home directory attribute'),
        'ldap_user_shell': _('Shell attribute'),
        'ldap_user_uuid': _('UUID attribute'),
        'ldap_user_objectsid': _("objectSID attribute"),
        'ldap_user_primary_group': _('Active Directory primary group attribute for ID-mapping'),
        'ldap_user_principal': _('User principal attribute (for Kerberos)'),
        'ldap_user_fullname': _('Full Name'),
        'ldap_user_member_of': _('memberOf attribute'),
        'ldap_user_modify_timestamp': _('Modification time attribute'),
        'ldap_user_shadow_last_change': _('shadowLastChange attribute'),
        'ldap_user_shadow_min': _('shadowMin attribute'),
        'ldap_user_shadow_max': _('shadowMax attribute'),
        'ldap_user_shadow_warning': _('shadowWarning attribute'),
        'ldap_user_shadow_inactive': _('shadowInactive attribute'),
        'ldap_user_shadow_expire': _('shadowExpire attribute'),
        'ldap_user_shadow_flag': _('shadowFlag attribute'),
        'ldap_user_authorized_service': _('Attribute listing authorized PAM services'),
        'ldap_user_authorized_host': _('Attribute listing authorized server hosts'),
        'ldap_user_authorized_rhost': _('Attribute listing authorized server rhosts'),
        'ldap_user_krb_last_pwd_change': _('krbLastPwdChange attribute'),
        'ldap_user_krb_password_expiration': _('krbPasswordExpiration attribute'),
        'ldap_pwd_attribute': _('Attribute indicating that server side password policies are active'),
        'ldap_user_ad_account_expires': _('accountExpires attribute of AD'),
        'ldap_user_ad_user_account_control': _('userAccountControl attribute of AD'),
        'ldap_ns_account_lock': _('nsAccountLock attribute'),
        'ldap_user_nds_login_disabled': _('loginDisabled attribute of NDS'),
        'ldap_user_nds_login_expiration_time': _('loginExpirationTime attribute of NDS'),
        'ldap_user_nds_login_allowed_time_map': _('loginAllowedTimeMap attribute of NDS'),
        'ldap_user_ssh_public_key': _('SSH public key attribute'),
        'ldap_user_auth_type': _('attribute listing allowed authentication types for a user'),
        'ldap_user_certificate': _('attribute containing the X509 certificate of the user'),
        'ldap_user_email': _('attribute containing the email address of the user'),
        'ldap_user_passkey': _('attribute containing the passkey mapping data of the user'),
        'ldap_user_extra_attrs': _('A list of extra attributes to download along with the user entry'),

        'ldap_group_search_base': _('Base DN for group lookups'),
        'ldap_group_object_class': _('Objectclass for groups'),
        'ldap_group_name': _('Group name'),
        'ldap_group_pwd': _('Group password'),
        'ldap_group_gid_number': _('GID attribute'),
        'ldap_group_member': _('Group member attribute'),
        'ldap_group_uuid': _('Group UUID attribute'),
        'ldap_group_objectsid': _("objectSID attribute"),
        'ldap_group_modify_timestamp': _('Modification time attribute for groups'),
        'ldap_group_type': _('Type of the group and other flags'),
        'ldap_group_external_member': _('The LDAP group external member attribute'),
        'ldap_group_nesting_level': _('Maximum nesting level SSSD will follow'),
        'ldap_group_search_filter': _('Filter for group lookups'),
        'ldap_group_search_scope': _('Scope of group lookups'),

        'ldap_netgroup_search_base': _('Base DN for netgroup lookups'),
        'ldap_netgroup_object_class': _('Objectclass for netgroups'),
        'ldap_netgroup_name': _('Netgroup name'),
        'ldap_netgroup_member': _('Netgroups members attribute'),
        'ldap_netgroup_triple': _('Netgroup triple attribute'),
        'ldap_netgroup_modify_timestamp': _('Modification time attribute for netgroups'),

        'ldap_service_search_base': _('Base DN for service lookups'),
        'ldap_service_object_class': _('Objectclass for services'),
        'ldap_service_name': _('Service name attribute'),
        'ldap_service_port': _('Service port attribute'),
        'ldap_service_proto': _('Service protocol attribute'),

        'ldap_idmap_range_min': _('Lower bound for ID-mapping'),
        'ldap_idmap_range_max': _('Upper bound for ID-mapping'),
        'ldap_idmap_range_size': _('Number of IDs for each slice when ID-mapping'),
        'ldap_idmap_autorid_compat': _('Use autorid-compatible algorithm for ID-mapping'),
        'ldap_idmap_default_domain': _('Name of the default domain for ID-mapping'),
        'ldap_idmap_default_domain_sid': _('SID of the default domain for ID-mapping'),
        'ldap_idmap_helper_table_size': _('Number of secondary slices'),

        'ldap_use_tokengroups': _('Whether to use Token-Groups'),
        'ldap_min_id': _('Set lower boundary for allowed IDs from the LDAP server'),
        'ldap_max_id': _('Set upper boundary for allowed IDs from the LDAP server'),
        'ldap_pwdlockout_dn': _('DN for ppolicy queries'),
        'wildcard_limit': _('How many maximum entries to fetch during a wildcard request'),
        'ldap_library_debug_level': _('Set libldap debug level'),

        # [provider/ldap/auth]
        'ldap_pwd_policy': _('Policy to evaluate the password expiration'),

        # [provider/ldap/access]
        'ldap_access_filter': _('LDAP filter to determine access privileges'),
        'ldap_account_expire_policy': _('Which attributes shall be used to evaluate if an account is expired'),
        'ldap_access_order': _('Which rules should be used to evaluate access control'),

        # [provider/ldap/chpass]
        'ldap_chpass_uri': _('URI of an LDAP server where password changes are allowed'),
        'ldap_chpass_backup_uri': _('URI of a backup LDAP server where password changes are allowed'),
        'ldap_chpass_dns_service_name': _('DNS service name for LDAP password change server'),
        'ldap_chpass_update_last_change': _('Whether to update the ldap_user_shadow_last_change attribute after a '
                                            'password change'),

        # [provider/ldap/sudo]
        'ldap_sudo_search_base': _('Base DN for sudo rules lookups'),
        'ldap_sudo_full_refresh_interval': _('Automatic full refresh period'),
        'ldap_sudo_smart_refresh_interval': _('Automatic smart refresh period'),
        'ldap_sudo_random_offset': _('Smart and full refresh random offset'),
        'ldap_sudo_use_host_filter': _('Whether to filter rules by hostname, IP addresses and network'),
        'ldap_sudo_hostnames': _('Hostnames and/or fully qualified domain names of this machine to filter sudo rules'),
        'ldap_sudo_ip': _('IPv4 or IPv6 addresses or network of this machine to filter sudo rules'),
        'ldap_sudo_include_netgroups': _('Whether to include rules that contains netgroup in host attribute'),
        'ldap_sudo_include_regexp': _('Whether to include rules that contains regular expression in host attribute'),
        'ldap_sudorule_object_class': _('Object class for sudo rules'),
        'ldap_sudorule_object_class_attr': _('Name of attribute that is used as object class for sudo rules'),
        'ldap_sudorule_name': _('Sudo rule name'),
        'ldap_sudorule_command': _('Sudo rule command attribute'),
        'ldap_sudorule_host': _('Sudo rule host attribute'),
        'ldap_sudorule_user': _('Sudo rule user attribute'),
        'ldap_sudorule_option': _('Sudo rule option attribute'),
        'ldap_sudorule_runas': _('Sudo rule runas attribute'),
        'ldap_sudorule_runasuser': _('Sudo rule runasuser attribute'),
        'ldap_sudorule_runasgroup': _('Sudo rule runasgroup attribute'),
        'ldap_sudorule_notbefore': _('Sudo rule notbefore attribute'),
        'ldap_sudorule_notafter': _('Sudo rule notafter attribute'),
        'ldap_sudorule_order': _('Sudo rule order attribute'),

        # [provider/ldap/autofs]
        'ldap_autofs_map_object_class': _('Object class for automounter maps'),
        'ldap_autofs_map_name': _('Automounter map name attribute'),
        'ldap_autofs_entry_object_class': _('Object class for automounter map entries'),
        'ldap_autofs_entry_key': _('Automounter map entry key attribute'),
        'ldap_autofs_entry_value': _('Automounter map entry value attribute'),
        'ldap_autofs_search_base': _('Base DN for automounter map lookups'),
        'ldap_autofs_map_master_name': _('The name of the automount master map in LDAP.'),

        # [provider/ldap/resolver]
        'ldap_iphost_search_base': _('Base DN for IP hosts lookups'),
        'ldap_iphost_object_class': _('Object class for IP hosts'),
        'ldap_iphost_name': _('IP host name attribute'),
        'ldap_iphost_number': _('IP host number (address) attribute'),
        'ldap_iphost_entry_usn': _('IP host entryUSN attribute'),
        'ldap_ipnetwork_search_base': _('Base DN for IP networks lookups'),
        'ldap_ipnetwork_object_class': _('Object class for IP networks'),
        'ldap_ipnetwork_name': _('IP network name attribute'),
        'ldap_ipnetwork_number': _('IP network number (address) attribute'),
        'ldap_ipnetwork_entry_usn': _('IP network entryUSN attribute'),

        # [provider/simple/access]
        'simple_allow_users': _('Comma separated list of allowed users'),
        'simple_deny_users': _('Comma separated list of prohibited users'),
        'simple_allow_groups': _('Comma separated list of groups that are allowed to log in. This applies only to '
                                 'groups within this SSSD domain. Local groups are not evaluated.'),
        'simple_deny_groups': _('Comma separated list of groups that are explicitly denied access. This applies only '
                                'to groups within this SSSD domain. Local groups are not evaluated.'),

        # [provider/proxy]
        'proxy_max_children': _('The number of preforked proxy children.'),

        # [provider/proxy/id]
        'proxy_lib_name': _('The name of the NSS library to use'),
        'proxy_resolver_lib_name': _('The name of the NSS library to use for hosts and networks lookups'),
        'proxy_fast_alias': _('Whether to look up canonical group name from cache if possible'),

        # [provider/proxy/auth]
        'proxy_pam_target': _('PAM stack to use'),

        # [provider/files]
        'passwd_files': _('Path of passwd file sources.'),
        'group_files': _('Path of group file sources.')
    }



import re
import logging

from sssd.source_files import Files
from sssd.source_journald import Journald

logger = logging.getLogger()


class Utils:

    def load(self, args):
        """
        Load the appropriate source reader.

        Args:
            args (Namespace): argparse parsed arguments

        Returns:
            Instantiated source object
        """
        if args.source == "journald":
            source = Journald()
        else:
            source = Files(args.logdir)
        return source

    def matched_line(self, source, patterns):
        """
        Yield lines which match any number of patterns (OR) in
        provided patterns list.

        Args:
            source (Reader): source Reader object
        Yields:
            lines matching the provided pattern(s)
        """
        for line in source:
            for pattern in patterns:
                re_obj = re.compile(pattern)
                if re_obj.search(line):
                    if line.startswith('   *  '):
                        continue
                    yield line



"""
Module with extensions to the pytest_multihost
"""

import time
import logging
import pytest
import pytest_multihost.config
import pytest_multihost.host
from .exceptions import SSSDException


class QeConfig(pytest_multihost.config.Config):
    """QeConfig subclass of multihost plugin to extend functionality."""

    extra_init_args = {'directory_manager', 'directory_password',
                       'rootdn', 'rootdn_pwd'}

    def __init__(self, **kwargs):
        self.log = self.get_logger(f'{__name__}.{type(self).__name__}')
        pytest_multihost.config.Config.__init__(self, **kwargs)

    def get_domain_class(self):
        """return custom domain class.

        This is needed to fully extend the config for custom
        multihost plugin extensions.

        Args:
            None

        Returns:
            None
        """
        return QeDomain

    def get_logger(self, name):
        """Override get_logger to set logging level.

        Args:
            name (str): Name of the logger

        Returns:
            log (obj): Logger object
        """
        log = logging.getLogger(name)
        log.propagate = False
        if not log.handlers:
            # set log Level
            log.setLevel(logging.DEBUG)
            handler = logging.StreamHandler()
            handler.setLevel(logging.DEBUG)
            # set formatter
            formatter = logging.Formatter(
                '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
            handler.setFormatter(formatter)
            log.addHandler(handler)
        return log

    def filter(self, descriptions):
        """
        Override default behavior to not filter hosts, so that it can work
        with dynamic topologies.
        """
        return


class QeBaseHost(pytest_multihost.host.BaseHost):
    """QeBaseHost subclass of multihost plugin BaseHost class."""


class QeHost(QeBaseHost):
    """QeHost subclass of multihost plugin host class.

    This extends functionality of the host class for SSSD QE purposes.
    Here we add support functions that will be very widely used across
    tests and must be run on any or all hosts in the environment.
    """

    @property
    def sys_hostname(self):
        """Get system hostname

        Args:
            None

        Returns:
            str: System hostname
        """
        cmd = self.run_command(['hostname'], raiseonerr=False)
        output = cmd.stdout_text.strip()
        if '\n' in output:
            ret = output.split('\n')
            name = (ret[len(ret) - 1])
            return name
        return cmd.stdout_text.strip()

    @property
    def fips(self):
        """ Check if system is fips enabled """
        fips_check_cmd = "fips-mode-setup --is-enabled"
        cmd = self.run_command(fips_check_cmd, raiseonerr=False)
        return bool(cmd.returncode != 0)

    @property
    def distro(self):
        """ Get contents of /etc/redhat-release

            :param None:
            :return str: contents of /etc/redhat-release
        """
        cmd = self.run_command(['cat', '/etc/redhat-release'],
                               raiseonerr=False)
        if cmd.returncode != 0:
            return 'Unknown Distro'
        return cmd.stdout_text.strip()

    def package_mgmt(self, package, action='install'):
        """ Install packages
            : param str package: Package name or list of packages
            : param str action: Install/uninstall/update
            : return str: Return code of the yum remove command
        """
        if 'Fedora' in self.distro or '8.' in self.distro or\
                '9.' in self.distro or '10.' in self.distro:
            pkg_cmd = 'dnf'
        else:
            pkg_cmd = 'yum'
        pkg_install_cmd = f'{pkg_cmd} -y {action} {package}'
        cmd = self.run_command(pkg_install_cmd, raiseonerr=False)
        if cmd.returncode != 0:
            print(f"{pkg_install_cmd} failed.\nOUT:{cmd.stdout_text}\nERR:{cmd.stderr_text}")
        return bool(cmd.returncode == 0)

    def service_sssd(self, action):
        """ Start/stop/restart sssd service based on RHEL Version
            :param str action: Action to be performed (start/stop/restart)
            :return: str Return code of the systemctl/service command
            :Exception Raises exception
        """
        # For Fedora, Atomic and RHELs 7, 8, 9 this should work.
        service_command = f'systemctl {action} sssd'
        if 'Red Hat Enterprise Linux' in self.distro and ' 6.' in self.distro:
            # RHEL 6 needs service command
            service_command = f"service sssd {action}"
        cmd = self.run_command(service_command, raiseonerr=False)
        if cmd.returncode == 0:
            time.sleep(10)
            return cmd.returncode
        self.run_command('journalctl -xeu sssd.service', raiseonerr=False)
        self.run_command('cat /var/log/sssd/sssd.log', raiseonerr=False)
        raise SSSDException(f'Unable to {action} sssd', 1)

    def yum_install(self, package):
        """ Install packages through yum

            :param str package: Name of the package to be installed
            :return str: Returncode of the yum command
            :Exception: None
        """
        cmd = self.run_command(['yum', '-y', 'install', package],
                               raiseonerr=False)
        return cmd.returncode

    def dnf_install(self, package):
        """ Install packages through dnf

            :param str package: Name of the package to be installed
            :return str: Returncode of the dnf command
            :Exception: None
        """
        install = f'dnf install -y --setopt=strict=0 {package}'
        cmd = self.run_command(install, raiseonerr=False)
        return cmd.returncode

    def yum_uninstall(self, package):
        """ Uninstall packages through yum
            :param str package: Name of the package to be uninstalled
            :return str: Return code of the yum remove command
            :Exception: None
        """

        cmd = self.run_command(['yum', '-y', 'remove', package],
                               raiseonerr=False)
        return cmd.returncode

    def dnf_uninstall(self, package):
        """ Uninstall packages through dnf
            :param str package: Name of the package to be uninstalled
            :return str: Return code of the dnf remove command
            :Exception: None
        """

        cmd = self.run_command(['dnf', '-y', 'remove', package],
                               raiseonerr=False)
        return cmd.returncode

    def detect_files_provider(self):
        """
        Detect the SSSD's files provider feature
        :param: None
        :return bool: Returns "True" if "files provider" is supported
        :Exception: None
        """
        check = f'ls "/usr/lib64/sssd/libsss_files.so"'
        cmd = self.run_command(check, raiseonerr=False)
        return cmd.returncode == 0


class QeWinHost(QeBaseHost, pytest_multihost.host.WinHost):
    """ Windows Host class

    Subclass of pytest_multihost.host.WinHost, QeBaseHost
    Functions defined provide extra attributes when using Windows AD

    Attributes:
        domainname (str): Return domainname of the AD machine
        domain_basedn_entry (str): Return AD basedn
        netbiosname (str): Rerurn the netbios name of the machine
        realm (str):  Return AD realm in upper case
        sys_hostname(str): Return full hostname of the machine
     """

    # These are defined as class properties to be overriden on the
    # instance level.
    _domainname = None
    _domain_basedn_entry = None
    _hostname = None
    _netbiosname = None
    _realm = None

    @property
    def domainname(self):
        """ Return Domain name """
        if self._domainname is None:
            cmd = self.run_command(
                ['domainname'], set_env=False, raiseonerr=False)
            self._domainname = cmd.stdout_text.strip()
        return self._domainname

    @property
    def sys_hostname(self):
        """ Return FQDN """
        if self._hostname is None:
            hostname = 'hostname -f'
            cmd = self.run_command(hostname, set_env=False, raiseonerr=False)
            self._hostname = cmd.stdout_text.strip().lower()
        return self._hostname

    @property
    def realm(self):
        """ Return AD Realm """
        if self._realm is None:
            self._realm = self.domainname.upper()
        return self._realm

    @property
    def domain_basedn_entry(self):
        """ Return base DN Entry of the """
        if self._domain_basedn_entry is None:
            domain_list = ['DC=' + string for string in
                           self.domainname.split('.')]
            list1 = map(str, domain_list)
            self._domain_basedn_entry = ','.join(list1)
        return self._domain_basedn_entry

    @property
    def netbiosname(self):
        """ Return netbios name """
        if self._netbiosname is None:
            cmd = "powershell.exe -inputformat none -noprofile "\
                  "'(Get-ADDomain -Current LocalComputer)'.NetBIOSName"
            self._netbiosname = self.run_command(cmd).stdout_text
        return self._netbiosname

    def _get_client_dn_entry(self, client):
        """ Return DN entry of client computer in AD """
        cmd = self.run_command(['dsquery', 'computer', '-name', str(client)],
                               set_env=False, raiseonerr=False)
        output = cmd.stdout_text.strip()
        return output

    def _get_user_dn_entry(self, user):
        """ Return DN entry of client computer in AD """
        cmd = self.run_command(['dsquery', 'user', '-name', user],
                               set_env=False, raiseonerr=False)
        output = cmd.stdout_text.strip()
        return output


class QeDomain(pytest_multihost.config.Domain):
    """ QeDomain subclass of multihost plugin domain class. """
    def __init__(self, config, name, domain_type):
        """
        Subclass of pytest_multihost.config.Domain

        :param obj config: config config
        :param str name: Name
        :param str domain_type:

        :return None:
        """
        # No need to call the super constructor as everything is done here
        # pylint: disable=super-init-not-called
        self.type = str(domain_type)
        self.config = config
        self.name = str(name)
        self.hosts = []

    host_classes = {'default': QeHost, 'windows': QeWinHost}



""" AD access control test ported from bash

:requirement: ad_access_control
:casecomponent: sssd
:subsystemteam: sst_idm_sssd
:upstream: yes
:status: approved
"""
import pytest
import re
import pexpect
import time
from sssd.testlib.common.utils import sssdTools
from sssd.testlib.common.expect import pexpect_ssh
from sssd.testlib.common.exceptions import SSHLoginException


def ssh_login(multihost, username):
    """
    Run ssh login for given username and return a log in status string, and output strings.
    :param str username: username for which ssh be executed
     return str status and str output
    :str status: 'failed, denied', or 'success'
    :str output could be 'login failed', or 'list of matched access denied pattern' or
     'output of id cmd executed after ssh'
    """
    client_hostname = multihost.client[0].sys_hostname
    client = pexpect_ssh(client_hostname, username, 'Secret123', debug=False)
    cmd = 'echo > /var/log/secure'
    multihost.client[0].run_command(cmd, raiseonerr=False)
    try:
        client.login()
    except SSHLoginException:
        return 'failed'
    except pexpect.EOF:
        time.sleep(1)
        log_str = multihost.client[0].get_file_contents('/var/log/secure').decode('utf-8')
        patt = re.compile(r'Access.*denied for user')
        if patt.search(log_str):
            return 'denied present'
        else:
            return 'denied missing'
    else:
        (stdout, _) = client.command(f'id {username}')
        client.logout()
        return 'Success'


@pytest.mark.usefixtures('joinad')
@pytest.mark.ad_access_control
@pytest.mark.tier2
class TestAccessControl(object):
    """ Test cases for BZ: 1268902
    :setup:
        1. Join to AD using realm command.
        2. Add the user using adcli user add.
    """
    @staticmethod
    def test_001_simple_allow_user_to_user1(multihost, create_aduser_group, backupsssdconf):
        """
        :title: Set simple_allow_user to user1
        :description: Allow only one ADuser to log in with simple_allow_user option
        :id: 31b37c8e-2ea4-45d3-90a6-b0be9deb1599
        :steps:
            1. Set 'simple_allow_users' to a AD-user
            2. Restart SSSD
            3. Log in with Allowed AD-user
            4. After login, execute 'id' command
        :expectedresults:
            1. Option should be correctly set
            2. SSSD should start correctly
            3. Allowed ADuser should be able to log in
            4. 'id' command should return user-information
        """
        tools = sssdTools(multihost.client[0])
        (aduser, _) = create_aduser_group
        domain_name = tools.get_domain_section_name()
        dom_section = f'domain/{domain_name}'
        sssd_params = {'access_provider': 'simple',
                       'simple_allow_users': f'{aduser}@{domain_name}'}
        tools.sssd_conf(dom_section, sssd_params, action='add')
        tools.clear_sssd_cache()
        ret = ssh_login(multihost, f'{aduser}@{domain_name}')
        ret1 = ssh_login(multihost, f'administrator@{domain_name}')
        assert ret == 'Success', 'allowed user failed to log in'
        assert ret1 == 'denied present', 'access denied log is absent in /var/log/secure'

    @staticmethod
    def test_002_too_much_logging_from_sssd_be(multihost, create_aduser_group, backupsssdconf):
        """
        :title: too much logging from sssd_be bz1269018
        :description: sssd_be is logging in the authentications in /var/log/messages
        :id: 82dd0c1a-cf62-4e15-978f-2a6e6c4c007c
        :steps:
            1. Clear /var/log/messages
            2. AD-user log in multiple times via ssh
        :expectedresults:
            1. /var/log/messages should be empty
            2. /var/log/messages should not have logs related sssd_be
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=1269018
        :customerscenario: Yes
        """
        (aduser, _) = create_aduser_group
        tools = sssdTools(multihost.client[0])
        domain_name = tools.get_domain_section_name()
        dom_section = f'domain/{domain_name}'
        sssd_params = {'access_provider': 'simple',
                       'simple_allow_users': f'{aduser}@{domain_name}'}
        tools.sssd_conf(dom_section, sssd_params, action='add')
        tools.clear_sssd_cache()
        multihost.client[0].run_command('echo > /var/log/messages', raiseonerr=False)
        client_hostname = multihost.client[0].sys_hostname
        for i in range(4):
            locals()[f'cl{i}'] = pexpect_ssh(client_hostname, f'{aduser}@{domain_name}', 'Secret123', debug=False)
            client = locals()[f'cl{i}']
            try:
                client.login()
            except SSHLoginException:
                pytest.fail(f'{aduser} failed to login')
            else:
                client.logout()
        log_str = multihost.client[0].get_file_contents('/var/log/messages').decode('utf-8')
        patt = re.compile(r'sssd_be')
        assert not patt.findall(log_str), 'The /var/log/messages is populated with sssd_be related logs'

    @staticmethod
    def test_003_simple_allow_user_to_dollar_symbol(multihost, create_aduser_group, backupsssdconf):
        """
        :title: Set simple_allow_user to dollar symbol
        :description: Setting simple_allow_user to '$' to deny all user log in
        :id: 3317cb9f-fa25-40b4-835b-d8c5da9cda5b
        :steps:
            1. Set simple_allow_user to $ symbol in sssd.conf and restart sssd
            2. Log in as AD user
        :expectedresults:
            1. Changes should reflected in restarted sssd
            2. Any AD-User log in should be denied
        """
        tools = sssdTools(multihost.client[0])
        (aduser, _) = create_aduser_group
        domain_name = tools.get_domain_section_name()
        dom_section = f'domain/{domain_name}'
        sssd_params = {'access_provider': 'simple',
                       'simple_allow_users': '$'}
        tools.sssd_conf(dom_section, sssd_params, action='add')
        tools.clear_sssd_cache()
        ret = ssh_login(multihost, f'{aduser}@{domain_name}')
        assert ret == 'denied present', 'Access denied log is absent in /var/log/secure'

    @staticmethod
    def test_simple_allow_user_to_invalid_user(multihost, create_aduser_group, backupsssdconf):
        """
        :title: Set simple_allow_user to an invalid user
        :id: fdde80ab-1364-4953-893e-62f7c3377ff8
        :description: Setting simple_allow_user to an invalid user, with this sssd would deny log in of all users
        :steps:
            1. Set simple_allow_user to non existing user in sssd.conf and restart sssd
            2. Log in as AD user
        :expectedresults:
            1. Changes should reflected in restarted sssd
            2. User log in should be denied
        """
        tools = sssdTools(multihost.client[0])
        (aduser, _) = create_aduser_group
        domain_name = tools.get_domain_section_name()
        dom_section = f'domain/{domain_name}'
        sssd_params = {'access_provider': 'simple',
                       'simple_allow_users': f'nonuser@{domain_name}'}
        tools.sssd_conf(dom_section, sssd_params, action='add')
        tools.clear_sssd_cache()
        ret = ssh_login(multihost, f'{aduser}@{domain_name}')
        assert ret == 'denied present', 'Access denied log is absent in /var/log/secure'

    @staticmethod
    def test_simple_deny_user_to_user1(multihost, create_aduser_group, backupsssdconf):
        """
        :title: Set simple_deny_user to a user
        :id: 8a068411-4ac7-4b15-b21e-210f3473d164
        :description: Setting simple_deny_user to an AD-user, with this sssd would deny log in of that user only
        :steps:
            1. Set simple_deny_user to a user in sssd.conf and restart sssd
            2. Log in as AD user
        :expectedresults:
            1. SSSD should load changes and start successfully
            2. User log in should be denied
        """
        tools = sssdTools(multihost.client[0])
        (aduser, _) = create_aduser_group
        domain_name = tools.get_domain_section_name()
        dom_section = f'domain/{domain_name}'
        sssd_params = {'access_provider': 'simple',
                       'simple_deny_users': f'{aduser}@{domain_name}'}
        tools.sssd_conf(dom_section, sssd_params, action='add')
        tools.clear_sssd_cache()
        ret = ssh_login(multihost, f'{aduser}@{domain_name}')
        ret1 = ssh_login(multihost, f'Administrator@{domain_name}')
        assert ret == 'denied present', 'Access denied log is absent in /var/log/secure'
        assert ret1 == 'Success', 'Allowed user failed to log in'

    @staticmethod
    def test_simple_deny_user_to_invalid_user(multihost, create_aduser_group, backupsssdconf):
        """
        :title: Set simple_deny_user to an invalid user
        :id: 76aea575-420b-4276-b0c6-383a4a8499a0
        :description: Setting simple_deny_user to an invalid user would allow all users to log in
        :steps:
            1. Set 'simple_deny_users' to an invalid AD-user
            2. Restart SSSD
            3. Log in with any AD-user
        :expectedresults:
            1. Option should be correctly set
            2. SSSD should start correctly
            3. Any ADuser should be able to log in
        """
        tools = sssdTools(multihost.client[0])
        (aduser, _) = create_aduser_group
        domain_name = tools.get_domain_section_name()
        dom_section = f'domain/{domain_name}'
        sssd_params = {'access_provider': 'simple',
                       'simple_deny_users': f'nonuser@{domain_name}'}
        tools.sssd_conf(dom_section, sssd_params, action='add')
        tools.clear_sssd_cache()
        ret = ssh_login(multihost, f'{aduser}@{domain_name}')
        assert ret == 'Success', 'ADuser log in failed'

    @staticmethod
    def test_simple_allow_groups_top_nested(multihost, create_aduser_group, create_nested_group, backupsssdconf):
        """
        :title: Set simple allow groups to the top-level nested group
        :id: ae19a604-1ce4-4a03-b1a6-4023aa71961f
        :description: Set simple_allow_groups to a top-level nested group. This top-level group will have one group
         as it's member. The member-group has one AD-user as group-member. The AD-user from group should
         be able to log in.
        :Steps:
            1. Create two groups.
            2. Add one group as a member of other group.
            3. Create an AD user and add that user as member of a member group
            4. Set 'simple_allow_groups' to a top level nested group
            5. Restart SSSD
            6. Log in with Allowed AD-user
        :expectedresults:
            1. Option should be correctly set
            2. SSSD should start correctly
            3. User and adding that to the group should succeed
            4. top-level-nested group is set as simple_allow_group
            5. SSSD Restart Should succeed
            6. Log in with Allowed AD-user Should succeed
        """
        run_id = create_nested_group
        aduser = f'testuser-{run_id}'
        l1_grp = f'testgrp-l1-{run_id}'
        tools = sssdTools(multihost.client[0], multihost.ad[0])
        domain_name = tools.get_domain_section_name()
        dom_section = f'domain/{domain_name}'
        sssd_params = {'access_provider': 'simple',
                       'simple_allow_groups': f'{l1_grp}@{domain_name}'}
        tools.sssd_conf(dom_section, sssd_params, action='add')
        tools.clear_sssd_cache()
        ret = ssh_login(multihost, f'{aduser}@{domain_name}')
        assert ret == 'Success', 'ADuser log in failed'

    @staticmethod
    def test_simple_deny_groups_top_nested(multihost, create_aduser_group, create_nested_group, backupsssdconf):
        """
        :title: Set simple deny groups to the top-level nested group
        :id: 2befaf3c-509d-4421-9276-a8328a9d48ec
        :description: Set simple_deny_groups to a top-level nested group. This top-level group will have one group
         as it's member. The member-group has one AD-user as group-member. The AD-user from group should be
         able to log in.
        :Steps:
        1. Create two groups.
        2. Add one group as a member of other group.
        3. Create an AD user and add that user as member of a member group
        4. Set 'simple_deny_groups' to a top level nested group
        5. Restart SSSD
        6. Log in with a denied nested group ADuser member
        :expectedresults:
        1. Option should be correctly set
        2. SSSD should start correctly
        3. User and adding that to the group should succeed
        4. top-level-nested group is set as simple_deny_group
        5. SSSD Restart Should succeed
        6. Log in of denied ADuser should be denied
        """
        run_id = create_nested_group
        aduser = f'testuser-{run_id}'
        l1_grp = f'testgrp-l1-{run_id}'
        tools = sssdTools(multihost.client[0], multihost.ad[0])
        domain_name = tools.get_domain_section_name()
        dom_section = f'domain/{domain_name}'
        sssd_params = {'access_provider': 'simple',
                       'simple_deny_groups': f'{l1_grp}@{domain_name}'}
        tools.sssd_conf(dom_section, sssd_params, action='add')
        tools.clear_sssd_cache()
        multihost.client[0].run_command(f'id {aduser}@{domain_name}', raiseonerr=False)
        ret = ssh_login(multihost, f'{aduser}@{domain_name}')
        assert ret == 'denied present', 'Access denied log is absent in /var/log/secure'

    @staticmethod
    def test_simple_allow_groups_invalid_group(multihost, create_aduser_group, backupsssdconf):
        """
        :title: Set simple allow groups to invalid group
        :id: d930ce65-0488-4e51-886a-c0ca4040fbaa
        :description: Set simple_allow_groups to a invalid group in sssd.conf.
         This should deny log in of any valid AD-user
        :steps:
            1.Set simple_allow_group to the non-existing group
            2.Log in as any AD-user
        :expectedresults:
            1. SSSD should restart correctly
            2. Log in should be denied
        """
        tools = sssdTools(multihost.client[0])
        (aduser, _) = create_aduser_group
        domain_name = tools.get_domain_section_name()
        dom_section = f'domain/{domain_name}'
        sssd_params = {'access_provider': 'simple',
                       'simple_allow_groups': f'nongrp@{domain_name}'}
        tools.sssd_conf(dom_section, sssd_params, action='add')
        tools.clear_sssd_cache()
        ret = ssh_login(multihost, f'{aduser}@{domain_name}')
        assert ret == 'denied present', 'Access denied log is absent in /var/log/secure'

    @staticmethod
    def test_simple_deny_groups_invalid_grp(multihost, create_aduser_group, backupsssdconf):
        """
        :title: Set simple deny groups to invalid group
        :id: 2befaf3c-509d-4421-9276-a8328a9d48ec
        :description: Set simple_deny_groups to a invalid group in sssd.conf.
         This should allow log in of any valid AD-user
        :steps:
            1.Set simple_deny_group to the non-existing group
            2.Log in as any AD-user
        :expectedresults:
            1. SSSD should restart correctly
            2. Log in should be successful
        """
        tools = sssdTools(multihost.client[0])
        (aduser, _) = create_aduser_group
        domain_name = tools.get_domain_section_name()
        dom_section = f'domain/{domain_name}'
        sssd_params = {'access_provider': 'simple',
                       'simple_deny_groups': f'invalidgroup@{domain_name}'}
        tools.sssd_conf(dom_section, sssd_params, action='add')
        tools.clear_sssd_cache()
        ret = ssh_login(multihost, f'{aduser}@{domain_name}')
        assert ret == 'Success', 'ADuser log in failed'

    @staticmethod
    def test_permit_all_users(multihost, create_aduser_group, backupsssdconf):
        """
        :title: Set access_provider to permit all users
        :id: 747a8268-69b9-4387-bce4-61e7facd03ad
        :description: Set access_provider to permit This should allow log in of any valid AD-user
        :steps:
            1.Set  access_provider to permit in sssd.conf
            2.Log in as any AD-user
        :expectedresults:
            1. SSSD should restart correctly
            2. Log in should be successful
        """
        tools = sssdTools(multihost.client[0])
        (aduser, adgrp) = create_aduser_group
        domain_name = tools.get_domain_section_name()
        dom_section = f'domain/{domain_name}'
        sssd_params = {'access_provider': 'permit'}
        tools.sssd_conf(dom_section, sssd_params)
        tools.clear_sssd_cache()
        ret = ssh_login(multihost, f'{aduser}@{domain_name}')
        assert ret == 'Success', 'ADuser log in failed'

    @staticmethod
    def test_deny_all_users(multihost, create_aduser_group, backupsssdconf):
        """
        :title: Set access provider to deny all users
        :id: 36f82023-3029-4fe4-82b1-7390f01bb5d6
        :description: Set access_provider to deny This should allow log in of any valid AD-user
        :steps:
            1.Set  access_provider to deny in sssd.conf
            2.Log in as any AD-user
        :expectedresults:
            1. SSSD should restart correctly
            2. Log in should be denied
        """
        (aduser, _) = create_aduser_group
        tools = sssdTools(multihost.client[0], multihost.ad[0])
        domain_name = tools.get_domain_section_name()
        multihost.client[0].run_command('realm deny --all', raiseonerr=False)
        tools.clear_sssd_cache()
        ret = ssh_login(multihost, f'{aduser}@{domain_name}')
        assert ret == 'denied present', 'Access denied log is absent in /var/log/secure'

    @staticmethod
    def test_dont_fail_auth_with_allow_rules(multihost, create_aduser_group, backupsssdconf):
        """
        :title: Set access_provider to permit all users
        :id: 4e746a3b-8f8e-4884-930f-2ae8ef72e626
        :description: When simple_allow_groups has one valid and one invalid group as inputs then
         sssd should ignore invalid group argument. SSSD should allow member-users from valid group
         to log in
        :steps:
            1. Define simple_allow_group with one valid group and one invalid group
            2. Log in from that valid-groups's member-user
        :expectedresults:
            1. SSSD should restart correctly
            2. Log in should be allowed
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=1170910
        :customerscenario: Yes
        """
        (aduser, adgroup) = create_aduser_group
        tools = sssdTools(multihost.client[0], multihost.ad[0])
        domain_name = tools.get_domain_section_name()
        dom_section = f'domain/{domain_name}'
        sssd_params = {'access_provider': 'simple',
                       'simple_allow_groups': f'{adgroup}@{domain_name}, non_grp@{domain_name}'}
        tools.sssd_conf(dom_section, sssd_params, action='add')
        tools.clear_sssd_cache()
        ret = ssh_login(multihost, f'{aduser}@{domain_name}')
        ret1 = ssh_login(multihost, f'Administrator@{domain_name}')
        assert ret == 'Success', 'allowed user failed to log in'
        assert ret1 == 'denied present', 'access denied log is absent in /var/log/secure'



""" IPA hbac test cases

:requirement: HBAC (ipa_provider)
:casecomponent: sssd
:subsystemteam: sst_idm_sssd
:upstream: yes
:status: approved
"""
import pytest
import time
import re
from sssd.testlib.ipa.utils import ipaTools
from sssd.testlib.common.utils import sssdTools
from sssd.testlib.common.expect import pexpect_ssh


@pytest.mark.usefixtures('default_ipa_users',
                         'disable_allow_all_hbac',
                         'reset_password')
@pytest.mark.tier2
@pytest.mark.hbac
class Testipahbac(object):

    def test_sssctl_sshd(self, multihost, hbac_sshd_rule):
        """
        :title: hbac: Verify using sssctl sshd service is allowed
        :id: 33a4288a-6c05-4472-a956-378736e030c8
        """
        sssctl_cmd = 'sssctl user-checks -s sshd foobar1'
        test_pam = re.compile(r'%s' % 'pam_acct_mgmt: Success')
        cmd = multihost.client[0].run_command(sssctl_cmd)
        result = test_pam.search(cmd.stderr_text)
        if not result:
            STATUS = 'FAIL'
        else:
            STATUS = 'PASS'
        assert STATUS == 'PASS'

    def test_hbac_changes(self, multihost, hbac_sshd_rule):
        """
        :title: hbac: verify hbac rule changes are applied
        :id: fcd4fc73-2425-46f8-8d66-95a49b5fb361
        """
        update_rule = "ipa hbacrule-remove-user --users='foobar1' test1"
        multihost.master[0].run_command(update_rule)
        sssctl_cmd = 'sssctl user-checks -s sshd foobar1'
        cmd = multihost.client[0].run_command(sssctl_cmd)
        test_pam = re.compile(r'%s' % 'pam_acct_mgmt: Permission denied')
        result = test_pam.search(cmd.stderr_text)
        if not result:
            STATUS = 'FAIL'
        else:
            STATUS = 'PASS'
        assert STATUS == 'PASS'

    def test_hbac_refresh_time(self, multihost):
        """
        :title: hbac: Verify cached hbac rule is applied
         for the refresh time period
        :id: c839fd33-65da-4252-82cf-5ba88ad02f55
        """
        ipa_server = ipaTools(multihost.master[0])
        ipa_client = ipaTools(multihost.client[0])
        sssd_client = sssdTools(multihost.client[0])
        domain_name = '%s/%s' % ('domain',
                                 sssd_client.get_domain_section_name())
        client_host = multihost.client[0].sys_hostname
        pexpect_ssh(client_host, 'foobar1', 'Secret123', debug=False)
        ipa_server.add_hbac_rule('test1', 'foobar1', client_host, 'sshd')
        multihost.client[0].service_sssd('stop')
        sssd_client.remove_sss_cache('/var/lib/sss/db')
        hbac_params = {'ipa_hbac_refresh': '60'}
        sssd_client.sssd_conf(domain_name, hbac_params)
        multihost.client[0].service_sssd('start')
        login_status = ipa_client.ssh_login('foobar1', 'Secret123',
                                            client_host, command='id')
        if login_status:
            status = 'PASS'
        # update the rule
        update_rule = "ipa hbacrule-remove-user --users='foobar1' test1"
        # sleep for 20 seconds
        time.sleep(20)
        multihost.master[0].run_command(update_rule)
        login_status = ipa_client.ssh_login('foobar1', 'Secret123',
                                            client_host, command='id')
        if login_status:
            status = 'PASS'
        time.sleep(45)
        # now it should not allow login
        login_status = ipa_client.ssh_login('foobar1', 'Secret123',
                                            client_host)
        if not login_status:
            status = 'PASS'
        sssd_client.sssd_conf(domain_name, hbac_params, action='delete')
        multihost.client[0].service_sssd('restart')
        ipa_server.del_hbac_rule('test1')
        assert status == 'PASS'

    def test_multiple_hbac_rules(self, multihost):
        """
        :title: hbac: Verify HBAC Evaluation happens per service
         when user is associated  with multiple hbac rules
        :id: 6981b637-f4ea-449b-8916-60cc938c4d0f
        """
        ipa_server = ipaTools(multihost.master[0])
        client_host = multihost.client[0].sys_hostname
        ipa_server.add_hbac_rule('test1', 'foobar1', client_host, 'sshd')
        ipa_server.add_hbac_rule('test2', 'foobar1', client_host, 'sudo')
        sssctl_cmd = 'sssctl user-checks -s sshd foobar1'
        test_pam = re.compile(r'%s' % 'pam_acct_mgmt: Success')
        cmd = multihost.client[0].run_command(sssctl_cmd)
        result = test_pam.search(cmd.stderr_text)
        if not result:
            STATUS = 'FAIL'
        else:
            STATUS = 'PASS'
        ipa_server.del_hbac_rule('test1')
        ipa_server.del_hbac_rule('test2')
        assert STATUS == 'PASS'

    def test_nested_groups(self, multihost):
        """
        :title: hbac: Verify hbac evaluation works as expected
         with nested group evaluation
        :id: fb2fd287-b217-487c-a59a-d827c426b0bb
        """
        ipa_server = ipaTools(multihost.master[0])
        client_host = multihost.client[0].sys_hostname
        groups = ['std_group', 'admin_group']
        for grp in groups:
            cmd = 'ipa group-add %s' % grp
            multihost.master[0].run_command(cmd)
        # Add members
        cmd1 = 'ipa group-add-member --users=foobar1 std_group'
        cmd2 = 'ipa group-add-member --users=foobar2 admin_group'
        multihost.master[0].run_command(cmd1, raiseonerr=False)
        multihost.master[0].run_command(cmd2, raiseonerr=False)
        # make admin_group member of std_group
        nested_group = 'ipa group-add-member --groups=admin_group std_group'
        multihost.master[0].run_command(nested_group, raiseonerr=False)
        # add rule
        ipa_server.add_hbac_rule('allow_ssh_access', 'std_group', client_host,
                                 'sshd', group=True)
        ipa_server.add_hbac_rule('allow_sudo_access', 'admin_group',
                                 client_host, 'sudo', group=True)
        users = ['foobar1', 'foobar2']
        for user in users:
            sssctl_cmd = 'sssctl user-checks -s sshd %s' % user
            cmd1 = multihost.client[0].run_command(sssctl_cmd)
            test_pam = re.compile(r'%s' % 'pam_acct_mgmt: Success')
            result = test_pam.search(cmd1.stderr_text)
            if not result:
                STATUS = 'FAIL'
            else:
                STATUS = 'PASS'
        ipa_server.del_hbac_rule('allow_ssh_access')
        ipa_server.del_hbac_rule('allow_sudo_access')
        for grp in groups:
            cmd = 'ipa group-del %s' % grp
            multihost.master[0].run_command(cmd)
        assert STATUS == 'PASS'

    def test_auto_private_group(self, multihost):
        """
        :title: hbac: Verify hbac rule associated with
         User private Groups
        :id: 99904ccd-bf2f-4c09-9636-92e036e19a0e
        """
        ipa_server = ipaTools(multihost.master[0])
        sssd_client = sssdTools(multihost.client[0])
        domain_name = '%s/%s' % ('domain',
                                 sssd_client.get_domain_section_name())
        client_host = multihost.client[0].sys_hostname
        pexpect_ssh(client_host, 'foobar1', 'Secret123', debug=False)
        multihost.client[0].service_sssd('stop')
        sssd_client.remove_sss_cache('/var/lib/sss/db')
        enable_pvtgroups = {'auto_private_groups': 'True'}
        sssd_client.sssd_conf(domain_name, enable_pvtgroups)
        multihost.client[0].service_sssd('start')
        cmd = 'ipa group-add std_group'
        multihost.master[0].run_command(cmd)
        # Add members
        cmd1 = 'ipa group-add-member --users=foobar1 std_group'
        multihost.master[0].run_command(cmd1, raiseonerr=False)
        # add rule
        ipa_server.add_hbac_rule('allow_ssh_access', 'std_group',
                                 client_host, 'sshd', group=True)
        sssctl_cmd = 'sssctl user-checks -s sshd foobar1'
        cmd1 = multihost.client[0].run_command(sssctl_cmd)
        test_pam = re.compile(r'%s' % 'pam_acct_mgmt: Success')
        result = test_pam.search(cmd1.stderr_text)
        if not result:
            STATUS = 'FAIL'
        else:
            STATUS = 'PASS'
        ipa_server.del_hbac_rule('allow_ssh_access')
        cmd = 'ipa group-del std_group'
        multihost.master[0].run_command(cmd)
        sssd_client.sssd_conf(domain_name, enable_pvtgroups, action='delete')
        multihost.client[0].service_sssd('restart')
        assert STATUS == 'PASS'



#!/usr/bin/env python
#  SSSD
#
#  SSSD python SSS API tests
#
#  Copyright (C) Red Hat
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function

import unittest
import sys
import os
import tempfile

BUILD_DIR = os.getenv('builddir') or "."
TEST_DIR = os.path.realpath(os.getenv('SSS_TEST_DIR') or ".")
MODPATH = tempfile.mkdtemp(prefix="tp_pysss_", dir=TEST_DIR)


class PysssImport(unittest.TestCase):
    def setUp(self):
        " Make sure we load the in-tree module "
        self.system_path = sys.path[:]
        sys.path = [MODPATH]

    def tearDown(self):
        " Restore the system path "
        sys.path = self.system_path

    def test_import(self):
        " Import the module and assert it comes from tree "
        try:
            dest_module_path = MODPATH + "/pysss.so"

            if sys.version_info[0] > 2:
                src_module_path = BUILD_DIR + "/.libs/_py3sss.so"
            else:
                src_module_path = BUILD_DIR + "/.libs/_py2sss.so"

            src_module_path = os.path.abspath(src_module_path)
            os.symlink(src_module_path, dest_module_path)

            import pysss
        except ImportError as ex:
            print("Could not load the pysss module. Please check if it is "
                  "compiled", file=sys.stderr)
            raise ex
        self.assertEqual(os.path.realpath(pysss.__file__),
                         os.path.realpath(MODPATH + "/pysss.so"))


class PysssEncryptTest(unittest.TestCase):
    def test_encrypt(self):
        obfuscator = pysss.password()

        val1 = obfuscator.encrypt("123", obfuscator.AES_256)
        self.assertEqual(len(val1), 96)

        val2 = obfuscator.encrypt("123", obfuscator.AES_256)
        self.assertEqual(len(val2), 96)

        self.assertNotEqual(val1, val2)


if __name__ == "__main__":
    error = 0

    suite = unittest.TestLoader().loadTestsFromTestCase(PysssImport)
    res = unittest.TextTestRunner().run(suite)
    if not res.wasSuccessful():
        error |= 0x1
        # need to bail out here because pysss could not be imported
        sys.exit(error)

    # import the pysss module into the global namespace, but make sure it's
    # the one in tree
    sys.path.insert(0, MODPATH)
    import pysss

    loadTestsFromTestCase = unittest.TestLoader().loadTestsFromTestCase

    suite = loadTestsFromTestCase(PysssEncryptTest)
    res = unittest.TextTestRunner().run(suite)
    if not res.wasSuccessful():
        error |= 0x2

    sys.exit(error)
